#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: dhcp-client.tcl,v 1.3 2002/11/15 13:23:38 joe Exp $
#
#
# The next line execs tclsh \
exec tclsh "$0" "$@"

# -- include pktsrc
package require pktsrc

# -- help/usage
proc help { name } {
  puts "$name \[option \[value]] ..."

  puts "-trace         Display protocol decodes"  
  puts "-packets       Display hex packet dumps"
  puts "-timestamp     Show timestamps in logging messages"
  puts "-interface     Set the physical interface to use, default is eth0"
  puts "-mac           Set the mac address"
  puts "-gui           Start in gui mode (requires Tk)"
  puts "-help          Display this help"
  puts ""

}

# -- set defaults
set options(interface) eth0
set options(myip)      0.0.0.0
set options(mymask)    0.0.0.0
set options(server)    0.0.0.0

proc draw_dhcp_gui { } {
   global stack1

   global config params pdsnIP myIP defaultMAC myMac do fieldWidth options
   global remotePPPIP localPPPIP stack1 dut

   # -- try to set the title and use this to check for X DISPLAY problems
   if [ catch { wm title . "dhcp tester" } reason ] {
      puts "ERROR: check DISPLAY variable"
      puts "It looks like the X connection failed"
      exit
   }
 
   set logo [ pktsrcFindLogo ]
   if { "$logo" != "" } {

      # -- add some control buttons
      frame .c -borderwidth 3 -relief groove
      pack configure .c -padx 2 -pady 2 -fill x

      set icon [ image create photo -file $logo ]
      canvas .c.icon2 -height 30
      .c.icon2 create image 20 15 -image $icon
      .c.icon2 create text 95 15 -text "DHCP Tester" -font { helvetica 18 bold }
      pack .c.icon2 -fill both -anchor nw
   }

   # -- create a label at the top
   label .l2 -text "DHCP options:" 
   pack configure .l2 -padx 5 -pady 1 -side top -anchor nw

   set i 0

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Interface" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable options(interface)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i 

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Client IP" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,clientIP)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left
 
   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Netmask" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,mask)
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Server IP" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,serverIP)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Default Gateway" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,gateway)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "DNS Server" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,dns)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Domain Name" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,domain)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Last xid" -anchor w
   entry .o${i}.e -relief sunken -width 16 -textvariable stack1(DHCP,xid)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Lease Obtained" -anchor w
   entry .o${i}.e -relief sunken -width 26 -textvariable stack1(DHCP,obtained)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left

   incr i

   frame .o$i
   pack configure .o$i -padx 15 -side top -anchor nw
   label .o${i}.l -width 20 -text "Lease Expires" -anchor w
   entry .o${i}.e -relief sunken -width 26 -textvariable stack1(DHCP,expires)
   .o${i}.e config -state disabled
   pack configure .o${i}.l .o${i}.e -padx 2 -side left


   # -- draw checkboxes for all the config options
   # -- create a label at the top
   label .l3 -text "Trace options:" 
   pack configure .l3 -padx 5 -pady 1 -side top -anchor nw

   set count 0
   foreach name "packets timestamp trace" {
      checkbutton .c$count -text $name \
         -variable options($name) -command dhcp_set_tracing
      pack configure .c$count -padx 25 -pady 1 -side top -anchor nw
      incr count
   }

   # -- add some control buttons
   frame .b -borderwidth 3 -relief groove -width "125"
   pack configure .b -padx 2 -pady 2 -fill x

   button .b.renew -relief groove -width 8 -text renew -command { dhcp_run }
   button .b.release -relief groove -width 8 -text release -command { dhcp_release }
   button .b.quit -relief groove -width 8 -text quit -command { exit }

   pack configure .b.renew .b.release .b.quit -padx 5 -pady 5 -side left

}

proc dhcp_set_tracing { } {
   global options stack1 port1

   # -- turn on verbose packet decoding
   if { $options(trace) == 1 } {
     set stack1(print) on
   } else {
     catch { unset stack1(print) }
   }

   # -- turn on packet timestamps
   if { $options(timestamp) == 1 } {
     set stack1(timestamp) on
   } else {
     catch { unset stack1(timestamp) }
   }

   # -- turn on raw packet dumps
   if { $options(packets) == 1 } {
      $port1 dump on
   } else {
      $port1 dump off
   }

}

proc dhcp_run { } {
   global stack1

   # -- start the DHCP client
   DHCP_client stack1
}

proc dhcp_release { } {
   global stack1

   # -- start the DHCP client
   DHCP_client_release stack1
}




#
# -- parse command line options
#

for { set i 0 } { $i < $argc } { incr i } {

  set arg [lindex $argv $i]

    switch -- $arg \
       "-trace" {
           set options(trace) on
           continue

       } "-packets" {
           set options(packets) 1
           continue

       } "-timestamp" {
           set options(timestamp) 1
           continue

       } "-version" {
           incr i
           set options(version) "[lindex $argv $i]"
           continue

       } "-interface" {
           incr i
           set options(interface) "[lindex $argv $i]"
           continue

       } "-gui" {
           set options(gui) 1
           continue

       } "-mac" {
           incr i
           set options(mac) "[lindex $argv $i]"
           continue

       } "-help" {
           help $argv0 
           exit

       }  default {
           puts "Unknown switch \"$arg\""
           exit
       }

    break
}


# -- open a packet source interface
set interface eth0
set myip $options(myip)
set mask $options(mymask)

if [ info exists options(interface) ] {
   set interface $options(interface)
}

if [ catch { set port1 [ pktsrc open $interface ] } reason ] {
  puts "Could not open interface $interface"
  puts "ERROR: $reason"
  help $argv0
  exit
}

# -- set up filters to only see dhcp and related traffic
$port1 filter "arp or port 67 or port 68 or icmp"
 
# -- turn on protocol decoding
if [ info exists options(trace) ] {
   set stack1(print) on
}

# -- turn on message timestamps
if [ info exists options(timestamp) ] {
   set stack1(timestamp) on
}

# -- turn on raw packet dumps
if [ info exists options(packets) ] {
   $port1 dump on
}  

# -- create stacks 
Stack_init stack1 $port1 $myip $mask DHCP

if [ info exists options(mac) ] {
   Stack_set_mac stack1 $options(mac)
}

if [ info exists options(gui) ] {

   if [ catch { package require Tk } reason ] {
      puts "ERROR: unable to load the Tk package"
      puts "ERROR: the -gui option requires Tk."
   }

   draw_dhcp_gui
   
} else {

   # -- start a DHCP client
   DHCP_client stack1

   # -- wait 5 seconds for all repsonses
   Event_wait 5000

   set newip [ IPv4_get stack1 address ] 

   if { "$newip" != "0.0.0.0" } {

     puts "DHCP client got new IP address $newip"

   } else {

     puts "DHCP client failed to get new IP address"
   }
}




