#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# ---------------------------------------------------------------------------
# $Id: flow-example.tcl,v 1.3 2002/07/15 15:29:53 joe Exp $
#
# flow example
#
# Send IP packets to various IP destination addresses
# ---------------------------------------------------------------------------
#
# The next line execs tclsh \
exec tclsh "$0" "$@"

# -- include pktsrc
package require pktsrc 1.1


# -- this is the base packet

set dstmac  "000000020304"
set srcmac  "000000010203"
set proto   "0800"
set header  "45"
set tos     "08"
set ip      "002ea5ea00003e"
set ipproto "11"
set cksum   "d2d2"
set ip_src  "1.1.1.1"
set ip_dst  "2.0.0.1"
set udp     "00010001001a0000"

# -- open a packet source handle
set handle [ pktsrc open eth0 ]

# -- loop through a bunch of IP addresses
for { set i 0 } { $i < 20 } { incr i } {

   set src [ iptohex $ip_src ]

   # -- nextip returns the next IP address - in pktsrc lib/net.tcl 
   set ip_dst [ nextip $ip_dst ]

   set dst [ iptohex $ip_dst ]

   # -- build the packet
   set packet [ format "%s%s%s%s%s%s%s%s%s%s%s" \
      $dstmac $srcmac $proto $header $tos $ip $ipproto $cksum $src $dst $udp ]

   $handle send-ip $packet 512
}


#
# -- repeat the same loop, but this time send the packets using an 802.1q tag

for { set i 0 } { $i < 20 } { incr i } {

   set src [ iptohex $ip_src ]

   # -- nextip returns the next IP address - in pktsrc lib/net.tcl 
   set ip_dst [ nextip $ip_dst ]

   set dst [ iptohex $ip_dst ]

   # -- build the packet
   #
   # 8100 -> 802.1q vlan
   # 00c8 -> vlan id 200

   set packet [ format "%s%s%s%s%s%s%s%s%s%s%s%s" \
      $dstmac $srcmac 810000c8 $proto $header $tos \
      $ip $ipproto $cksum $src $dst $udp ]

   # -- send-ip will detect the vlan tag and still compute the correct checksum
   $handle send-ip $packet 512
}














