#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ospf.tcl,v 1.6 2002/09/24 10:30:06 joe Exp $
#
#
# The next line execs tclsh \
exec tclsh "$0" "$@"

# 
# Simple host emulation
#
# The following example creates 1 instances of the pktsrc stack
# and starts OSPF on that node.

# -- include pktsrc
package require pktsrc 1.1

# -- open a packet source interface

set port1 [ pktsrc open eth0 ] 

# -- place filters on each packet source
$port1 filter "arp or host 192.168.1.15 or proto 89"


# -- create stacks 

Stack_init stack1 $port1 192.168.1.15 255.255.255.0
 
# -- change the default TTL for one stack
set stack1(IPv4,ttl) 1a

# -- name each stack to improve logging
Stack_name stack1 host-A

set stack1(print) on
set stack1(timestamp) on 

# -- turn on packet dump at the hex level
#$port1 dump on

# -- start the hello protocol
OSPF_init_hello stack1

# -- set up OSPF authentication

# -- example for simple authentication
# OSPF_set_auth stack1 simple qacafe

# -- example for md5 authentication
# OSPF_set_auth stack1 md5 qacafe 1

# -- wait for an adjaceny to occur
if [ OSPF_wait_state stack1 [iptohex $dut] Full 60 ] {
   puts "OK, we have an adjacency with $dut"
} else {
   puts "Failed to establish adjacency after 60 seconds and gave up!"
   exit
}


# -- enter the main event log
vwait forever


