#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ping.tcl,v 1.5 2002/07/17 02:19:43 joe Exp $
#
#
# The next line execs tclsh \
exec tclsh "$0" "$@"

# 
# Simple ping implementation
#
# This is a simple ping client that demonstrates some of the basic
# pktsrc stack mechanisms. It shows adding a callback and handling
# async packets.
#

# -- include pktsrc
package require pktsrc 1.1

# -- open a packet source interface

set port1 [ pktsrc open eth0 ] 

set host 192.168.1.25

# -- place filters on the packet source to only receive my packets (or arp)
$port1 filter "arp or host $host"


# -- create stack
Stack_init stack1 $port1 $host 255.255.255.0
 
# -- name each stack to improve logging
Stack_name stack1 ping-A

# -- uncomment if you want to see dumps of all the packets
set stack1(print) on
#set stack1(trace) on

# -- turn on packet dump at the hex level
#$port1 dump on

# -- here is the DUT for this example
set dut 192.168.1.11

# -- Add a callback for ping
Stack_add_callback stack1 icmpType 00 ping_response

# -- set the default TOS byte on this stack
set stack1(IPv4,tos) 7e

IPv4_set stack1 default 192.168.1.1
IPv4_set stack1 mtu 52

#
# -- ping response callback

proc ping_response { packet stack options } {
   upvar $stack s

   # -- get the contents of the IP packet from the stack
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the contents of the ICMP packet from the stack
   set icmpState $s(current,ICMP)
   upvar #$s(current,ICMP,level) $icmpState icmp 

   # -- get the length of the IP packet
   set len [expr 0x$ip(length)]

   # -- display the result

   # -- demonstrate packet timestamps
   puts -nonewline "$s(current,timestampSec).$s(current,timestampUsec) "

   # -- now display the usual
   puts "$len bytes from [hextoip $ip(src)]: icmp_seq=[expr 0x$icmp(sequence)]"
}


#
# -- send 100 pings

set data [ string repeat "feed" 80 ]


for { set i 1 } { $i <= 100 } { incr i } {

   # -- build an ICMP echo request
   set packet [ ICMP_build_echo_request stack1 01 $i $data ]
 
   # -- send the packet
   ICMP_send stack1 $packet $dut

   # -- wait 1 second
   Event_wait 1000

}








