#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: tcp-client.tcl,v 1.3 2002/07/17 02:19:43 joe Exp $
#
#
# TCP client example
#

 
# The next line execs tclsh \
exec tclsh "$0" "$@"

# -- include pktsrc
package require pktsrc

# -- open a packet source interface
set port1 [ pktsrc open eth0 ] 

# -- set my IP address on this interface (should be unused address)
set host 192.168.1.25

# -- place filters on the packet source to only receive my packets (or arp)
$port1 filter "arp or host $host"

# -- now create the stack
Stack_init stack1 $port1 $host 255.255.255.0
 
# -- name each stack to improve logging
Stack_name stack1 tcp-example

# -- uncomment if you want to see dumps of all the packets
set stack1(print) on

# -- turn on timestamps
set stack1(timestamp) on

# -- turn on packet dump at the hex level
#$port1 dump on

# -- add default route
IPv4_set stack1 default 192.168.1.150

# -- open a connection to www.google.com
set state [ TCP_open stack1 tcp 216.239.35.100 80 ]

puts "# The tcp session is in state '$state'"

#
# -- if the session is open send a request
if { "$state" == "established" } {

   # -- send an HTTP get request on this tcp connection
   set bytes [ HTTP_send stack1 tcp "GET /index.html HTTP/1.0" ]
   puts "# sent $bytes bytes"


   # -- read the response
   set bytes [ HTTP_read stack1 tcp page ]

   puts "# received $bytes bytes"
   if { "$bytes" > 0 } {
       puts "$page"
   }   

   # -- close connection
   TCP_close stack1 tcp
}

# -- wait for close to finish (http server probably closed connection)
Event_wait 1000







