#!/bin/sh
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: vnode-server.tcl,v 1.1 2002/11/03 13:29:00 joe Exp $
#
#
# Vnode server example
#
#
# This script gives an example of vnode usage. It creates 1000 vnodes that
# can be reached through a real stack. Basically, the real stack acks like
# an IPv4 gateway for all the vnode addresses.
#
 
# The next line execs tclsh \
exec tclsh "$0" "$@"

# -- include pktsrc
package require pktsrc

# -- open a packet source interface
set port1 [ pktsrc open eth0 ] 

# -- set my IP address on this interface (should be unused address)
set host 192.168.1.25

# -- now create the stack
Stack_init stack1 $port1 $host 255.255.255.0

# -- use a fake mac address for this stack
Stack_set_mac stack1 000203040506

# -- place filters on the packet source to only receive my packets (or arp)
$port1 filter "arp or host $host or ether dst [hextomac $stack1(mac)]"

# -- name each stack to improve logging
Stack_name stack1 vnode-example

# -- uncomment if you want to see dumps of all the packets
set stack1(print) on

# -- turn on timestamps
set stack1(timestamp) on

# -- turn on packet dump at the hex level
#$port1 dump on

# -- add default route
IPv4_set stack1 default 192.168.1.100
Event_wait 1000


# -- now create 1000 vnodes that will be reachable through 192.168.1.25
set startip 10.1.1.1

for { set i 0 } { $i < 1000 } { incr i } {

   Stack_init vnode$i $stack1(port) $startip 255.255.255.255 vnode$i -1 1
   Stack_add_vnode stack1 vnode$i

   # -- start a dummy server on port 80 for this vnode
   HTTP_dummy_server vnode$i 80

   # -- bump the next ip address
   set startip [ nextip $startip ]
}

# -- just wait for ever accepting connections
vwait forever

#
# -- now if you point a static route for 10.0.0.0/8 to 192.168.1.25,
# you should be able to connect to each vnode through your web
# browser.
#






