#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: dhcp.tcl,v 1.23 2002/12/12 14:06:36 joe Exp $
#
#
# pktsrc DHCP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

#
# Set the DHCP filter
#

proc DHCP_set_standard_filter { stack } {
   upvar $stack s

   set port $s(port)

   # -- set the standard filter to receive only DHCP related traffic
   $port filter "$s(DHCP,standard-filter)"
}

#
# Get the DHCP filter
#

proc DHCP_get_standard_filter { stack } {
   upvar $stack s

   return "$s(DHCP,standard-filter)"
}

proc DHCP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {
      DHCP_decode $packet dhcp
      DHCP_print dhcp
      return
   }

   incr s(DHCP,inpackets)

   # -- remember this packet in the 'dhcp' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s DHCP dhcp ]
   
   # -- now decode the packet into the 'dhcp' variable
   DHCP_decode $packet dhcp

   # -- check for callbacks at the Request/Reply level

   if [ info exists s(dhcpType,$dhcp(op)) ] {

      # -- drive any input callbacks
      foreach c $s(dhcpType,$dhcp(op),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }

   }

   # -- now check for packets that Option type 53 level
   set type [ DHCP_find_option dhcp 53 ]

   if [ info exists s(dhcp53Type,$type) ] {

      # -- drive any input callbacks
      foreach c $s(dhcp53Type,$type,callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   }
}

proc DHCP_reply_input { packet stack options } {
   upvar $stack s

   # -- get the contents of the IP packet from the stack
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the contents of the UDP packet from the stack
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp
 
   # -- get the contents of the DHCP packet from the stack
   set dhcpState $s(current,DHCP)
   upvar #$s(current,DHCP,level) $dhcpState dhcp
 
   set type [ DHCP_find_option dhcp 53 ]
   set address  $dhcp(yiaddr)
   set server [ DHCP_find_option dhcp 54 ]



   
   if { "$type" != "" } {
     pktsrcInfo s "Received [ DHCP_get_message_type $type ], xid $dhcp(xid)"

     if [ info exists s(DHCP,client,xid,discovery,$dhcp(xid)) ] {

       if { "$type" == "02" } {
          set sip [ hextoip $server ]
          set addr [ hextoip $address ]
          pktsrcInfo s "Server $sip offered IP address $addr"


          DHCP_send_request s $dhcp(xid) $sip $addr 
       } elseif { "$type" == "05" } {

          set sip [ hextoip $server ]
          set mask   [ DHCP_find_option dhcp 01 ]
          set seconds   [ DHCP_find_option dhcp 51 ]
          set domain   [ DHCP_find_option dhcp 15 ]

          # -- find any dns servers specified
          DHCP_find_dns_option dhcp dns1 dns2 dns3

          set gateway   [ DHCP_find_option dhcp 3 ]
          set newIp [ hextoip $address ]



	  if { "$mask" != "" } {
              set newMask [ hextoip $mask ]

    	      if { $domain != "" } {
                 set name [ pktsrc ascii2text $domain ]
                 set s(DHCP,domain) $name
              } else {
                 set s(DHCP,domain) ""
              }

              if { "$dns1" != "0.0.0.0" } {
                  set s(DHCP,dns)   $dns1
                  set s(DHCP,dns2)  $dns2
                  set s(DHCP,dns3)  $dns3

              } else {
                  set s(DHCP,dns)   0.0.0.0
                  set s(DHCP,dns2)  0.0.0.0
                  set s(DHCP,dns3)  0.0.0.0
              }

              if { "$gateway" != "" } {
                  set s(DHCP,gateway)   [ hextoip $gateway ]
              } else {
                  set s(DHCP,gateway)   0.0.0.0
              }


              set s(DHCP,clientIP) $newIp
              set s(DHCP,mask)     $newMask
              set s(DHCP,serverIP)   $sip  

              # -- calculate lease times for display purposes
              set s(DHCP,obtained) [ clock format [ clock seconds ]]
              set s(DHCP,expires) \
                 [ clock format [ expr 0x$seconds + [ clock seconds ]]]

              # -- okay I got a DHCPACK, set the address
              pktsrcInfo s "Address accepted, configure IP with $newIp/$newMask, gateway $s(DHCP,gateway)"
              pktsrcInfo s \
                 "Received DNS server $s(DHCP,dns), domain '$s(DHCP,domain)'"

              IPv4_reconfigure s $newIp $newMask
              IPv4_set s default $s(DHCP,gateway)
              
          } else {
              pktsrcWarn s "No netmask option found"
          }              
       }
     } else {
       pktsrcInfo s "Ignoring unknown xid $dhcp(xid)"
     }
   }

}

proc DHCP_request_input { packet stack options } {
   upvar $stack s


   # -- get the contents of the IP packet from the stack
   set ipState $s(current,IPv4)
   upvar #$s(current,IPv4,level) $ipState ip

   # -- get the contents of the UDP packet from the stack
   set udpState $s(current,UDP)
   upvar #$s(current,UDP,level) $udpState udp
 
   # -- get the contents of the DHCP packet from the stack
   set dhcpState $s(current,DHCP)
   upvar #$s(current,DHCP,level) $dhcpState dhcp
 
   set type [ DHCP_find_option dhcp 53 ]
   # set address  $dhcp(yiaddr)
   # set server [ DHCP_find_option dhcp 54 ]

   # -- remember the giaddr if one is set   
   set s(DHCP,relayIp) [ hextoip $dhcp(giaddr) ]
   
   if { "$type" != "" } {

     pktsrcInfo s "Received [ DHCP_get_message_type $type ], xid $dhcp(xid)"

     if { $s(DHCP,server,mode) == "off" } {
       return
     }

       if { "$type" == "01" } {

          # -- got discovery 

          # -- get the mac address
          set mac [ string range $dhcp(chaddr) 0 11 ]

          pktsrcInfo s \
             "DHCP client discovery from [hextomac $mac] with xid $dhcp(xid)"

          # -- send back offer with my IP as the server
          DHCP_send_offer s $dhcp(xid) $mac [ hextoip $s(ip) ] [ hextoip $ip(src) ]

       } elseif { "$type" == "03" } {

          # -- got request 

          # -- get the mac address
          set mac [ string range $dhcp(chaddr) 0 11 ]

          pktsrcInfo s \
             "DHCP client request from [hextomac $mac] for [ hextoip $dhcp(ciaddr) ] (xid $dhcp(xid))"

          # -- I should do some verification here
          set requestIp [ hextoip $dhcp(ciaddr) ]

          # -- should I use ciaddr or the requested address ?
          set option50 [ DHCP_find_option dhcp 50 ]

	  if { "$option50" != "" } {
              set rip [ hextoip $option50 ]
              pktsrcInfo s "Requested IP is $rip"
              set requestIp $rip
 
              # if { "$rip" != "$requestIp" } {
              #   pktsrcWarn s "requested ip does not match ciaddr -> $requestIp"
              # } 
          } 

	  if { "$requestIp" != "0.0.0.0" && \
               "$requestIp" != "$s(DHCP,server,address)" } {

              # -- this request is not for the correct address, send NAK
              DHCP_send_nack s $dhcp(xid) $mac [ hextoip $s(ip) ] [ hextoip $ip(src) ]

          } else {

              # -- send back offer with my IP as the server
              DHCP_send_ack s $dhcp(xid) $mac [ hextoip $s(ip) ] [ hextoip $ip(src) ]
          }

          # -- start timing out the lease
          set endTime [ expr [ clock seconds ] + $s(DHCP,server,lease) ]
          set s(DHCP,server,leaseEnd) $endTime 
       }
   }

}

#
#
# DHCP_init - set up any deaults for DHCP
#

proc DHCP_init { stack } {
   upvar $stack s

   set s(DHCP,clientIP) 0.0.0.0
   set s(DHCP,serverIP) 0.0.0.0
   set s(DHCP,yourIP)   0.0.0.0
   set s(DHCP,mask)     0.0.0.0
   set s(DHCP,xid)      0x0
   set s(DHCP,obtained) ""
   set s(DHCP,expires)  ""
   set s(DHCP,gateway)  0.0.0.0
   set s(DHCP,dns)      0.0.0.0
   set s(DHCP,dns2)     0.0.0.0
   set s(DHCP,dns3)     0.0.0.0
   set s(DHCP,dns4)     0.0.0.0
   set s(DHCP,domain)   ""
   set s(DHCP,standard-filter)  "arp or port 67 or port 68 or icmp"
   set s(DHCP,relayIp)   0.0.0.0

   # -- server related
   set s(DHCP,server,clientMac) 000000000000
   set s(DHCP,server,leaseEnd)  [ clock seconds ]
   set s(DHCP,site_option_count) 0
   
   # -- broadcast mode - off by default
   set s(DHCP,broadcast-bit) 0

   # -- set the initial packet counters
   set s(DHCP,inpackets) 0

   # -- the default hostname is the stack name

   set s(DHCP,hostname) $s(name)

   if { "$s(DHCP,hostname)" == "" } {
       set s(DHCP,hostname) fred
   }
}


proc DHCP_decode { packet DHCP } {
   upvar $DHCP out

#   0                   1                   2                   3
#   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |     op (1)    |   htype (1)   |   hlen (1)    |   hops (1)    |
#   +---------------+---------------+---------------+---------------+
#   |                            xid (4)                            |
#   +-------------------------------+-------------------------------+
#   |           secs (2)            |           flags (2)           |
#   +-------------------------------+-------------------------------+
#   |                          ciaddr  (4)                          |
#   +---------------------------------------------------------------+
#   |                          yiaddr  (4)                          |
#   +---------------------------------------------------------------+
#   |                          siaddr  (4)                          |
#   +---------------------------------------------------------------+
#   |                          giaddr  (4)                          |
#   +---------------------------------------------------------------+
#   |                                                               |
#   |                          chaddr  (16)                         |
#   |                                                               |
#   |                                                               |
#   +---------------------------------------------------------------+
#   |                                                               |
#   |                          sname   (64)                         |
#   +---------------------------------------------------------------+
#   |                                                               |
#   |                          file    (128)                        |
#   +---------------------------------------------------------------+
#   |                                                               |
#   |                          options (variable)                   |
#   +---------------------------------------------------------------+
#

   # -- parse out all the fields we find interesting ...

   set out(op)             [ string range $packet 0 1 ]
   set out(htype)          [ string range $packet 2 3 ]
   set out(hlen)           [ string range $packet 4 5 ]
   set out(hops)           [ string range $packet 6 7 ]
   set out(xid)            [ string range $packet 8 15 ]
   set out(secs)           [ string range $packet 16 19 ]
   set out(flags)          [ string range $packet 20 23 ]

   set out(ciaddr)         [ string range $packet 24 31 ]
   set out(yiaddr)         [ string range $packet 32 39 ]
   set out(siaddr)         [ string range $packet 40 47 ]
   set out(giaddr)         [ string range $packet 48 55 ]

   set out(chaddr)         [ string range $packet 56 87 ]
   set out(sname)          [ string range $packet 88 215 ]
   set out(file)           [ string range $packet 216 471 ]
   set out(magic-cookie)   [ string range $packet 472 479 ]

   set out(options) 0


   # -- check for options to decode   
   set len [ string length $packet ]

   if { $len > 480 } {
      set opStart [ string range $packet 480 end ]
      DHCP_decode_options $opStart out

      # -- check for overload option, we may have more options
      set type [ DHCP_find_option out 52 ]

      if { "$type" == "03" } {
        DHCP_decode_options "$out(sname)$out(file)" out
      } elseif { "$type" == "02" } {
        DHCP_decode_options "$out(sname)" out
      } elseif { "$type" == "01" } {
        DHCP_decode_options "$out(file)" out
      }
   }
}

proc DHCP_decode_options { packet dhcp } {
   upvar $dhcp out

   # -- set the count to the current number of options
   set count $out(options)

   # -- options should start at the beginning of the buffer
   set p 0

   while { 1 } {
      
      set tag [ string range $packet $p [expr $p + 1]]
      # puts "tag is $tag"

      if { $tag == "00" } {
        # -- pad option
        incr p 2
      } elseif { $tag == "ff" } {
        # -- end option
        break
      }

      incr p 2
      set length [ string range $packet $p [expr $p + 1]]
      # puts "length is $length"

      incr p 2

      set out(option,$count,tag) $tag
      set out(option,$count,length) $length

      # -- handle 0 length option
      if { $length == 0 } {
          set out(option,$count,data) ""
      } else {
          set out(option,$count,data) \
            [ string range $packet $p [expr $p + 2 * 0x$length - 1 ]]
          incr p [ expr 0x$length * 2  ]
      }

      # puts "data is $out(option,$count,data)"
      incr count
   } 

   # -- set the total number of options found, excluding 'pad' and 'end' options
   set out(options) $count  
}



proc DHCP_print { DHCP } {
   upvar $DHCP out

   puts "DHCP: ---- DHCP Packet ----"
   puts "DHCP: "
   puts "DHCP: Type                      = [DHCP_op_type $out(op)] (0x$out(op))"
   puts "DHCP: Hardware Type             = 0x$out(htype)"
   puts "DHCP: Hardware Address Len      = 0x$out(hlen)"
   puts "DHCP: Hops                      = 0x$out(hops)"
   puts "DHCP: Transaction ID            = 0x$out(xid) (xid)"

   puts "DHCP: Seconds                   = 0x$out(secs)"

   if { "$out(flags)" == "8000" } { 
      set broadcast "(BROADCAST BIT)"
   } else {
      set broadcast ""
   }

   puts "DHCP: Flags                     = 0x$out(flags) $broadcast"

   puts "DHCP: Client IP                 = [ hextoip $out(ciaddr)] (ciaddr)"
   puts "DHCP: Your IP                   = [ hextoip $out(yiaddr)] (yiaddr)"
   puts "DHCP: Server IP                 = [ hextoip $out(siaddr)] (siaddr)"
   puts "DHCP: Relay IP                  = [ hextoip $out(giaddr)] (giaddr)"
   puts "DHCP: Client HW Addr            = [ hextomac [ string range $out(chaddr) 0 11 ]] (chaddr)"

   puts "DHCP: ---- Options ----"
   puts "DHCP:"

   if { $out(options) > 0 } {
      puts "DHCP: Magic Cookie              = [ hextoip $out(magic-cookie)]"
      DHCP_print_options out
   } else {
      puts "DHCP: No options found"
   }

   puts "DHCP:"
   puts "DHCP: ---- End DHCP Packet ----"
}

proc DHCP_print_options { dhcp } {
   upvar $dhcp out


   for { set i 0 } { $i < $out(options) } { incr i } {

      # -- get the option tag in decimal
      set tag [ expr 0x$out(option,$i,tag) ]
      set length $out(option,$i,length)
      set data $out(option,$i,data)

      if { "$tag" == "01" } {

         # -- subnet mask
         puts "DHCP: Mask                      = [ hextoip $data]"

      } elseif { "$tag" == "03" } {

         # -- router option
         puts "DHCP: Routers"
         set rCount [ expr 0x$length / 4 ]
	 for { set j 1 } { $j <= $rCount } { incr j } {
             set r [ string range $data [ expr ($j-1)*8 ] [ expr ($j-1) * 8 + 7 ]] 
             puts "DHCP:   Router $j                = [ hextoip $r]"
         }

      } elseif { "$tag" == "04" } {

         # -- time server option
         set rCount [ expr 0x$length / 4 ]
         puts "DHCP: Time Servers"
	 for { set j 1 } { $j <= $rCount } { incr j } {
             set r [ string range $data [ expr ($j-1)* 8 ] [ expr ($j-1) * 8 + 7 ]] 
             puts "DHCP:   Time Server $j           = [ hextoip $r]"
         }

      } elseif { "$tag" == "06" } {

         # -- DNS server option
         set rCount [ expr 0x$length / 4 ]
         puts "DHCP: DNS Servers"
	 for { set j 1 } { $j <= $rCount } { incr j } {
             set r [ string range $data [ expr ($j-1)* 8 ] [ expr ($j-1) * 8 + 7 ]] 
             puts "DHCP:   DNS Server $j            = [ hextoip $r]"
         }
      } elseif { "$tag" == "12" } {

         # -- Host Name Option
         set name [ pktsrc ascii2text $data ]

         puts "DHCP: Host Name                 = $name"

      } elseif { "$tag" == "15" } {

         # -- Domain Name Option
         set name [ pktsrc ascii2text $data ]

         puts "DHCP: Domain Name               = $name"

      } elseif { "$tag" == "26" } {

         # -- Domain Name Option
         set mtu [ expr 0x$data ]

         puts "DHCP: Interface MTU             = $mtu"

      } elseif { "$tag" == "50" } {
  
         puts "DHCP: Requested IP Address      = [ hextoip $data]"
        
      } elseif { "$tag" == "51" } {

         set hours [ expr 0x$data / 3600.0 ]
         puts "DHCP: IP Address Lease Time     = [expr 0x$data] seconds ($hours hours)"

      } elseif { "$tag" == "52" } {

         set olStr ""
	 if { "$data" == "01" } {
             set olStr file
         } elseif { "$data" == "02" } {
             set olStr sname
         } elseif { "$data" == "03" } {
             set olStr "both file and sname"
         } else {
             set olStr "unknown use"
         }
         puts "DHCP: Option Overload           = $olStr (0x$data)"

      } elseif { "$tag" == "53" } {

         puts "DHCP: Message Type              = [ DHCP_get_message_type $data]"

      } elseif { "$tag" == "54" } {

         puts "DHCP: Server Identifier         = [ hextoip $data]"

      } elseif { "$tag" == "55" } {

         puts "DHCP: Parameter Request List"
         puts "DHCP:  Length                   = [expr 0x$length]"
         DHCP_print_parameter_list $data [expr 0x$length]

      } elseif { "$tag" == "57" } {

         puts "DHCP: Maximum Message Size      = [ expr 0x$data]"

      } elseif { "$tag" == "58" } {

         puts "DHCP: Renewal Time (T1)         = [ expr 0x$data] seconds"

      } elseif { "$tag" == "59" } {

         puts "DHCP: Rebinding Time (T2)       = [ expr 0x$data] seconds"

      } elseif { "$tag" == "61" } {

         puts "DHCP: Client ID                 = $data"

      } else {

         # -- default option display

         puts "DHCP: Option Type               = [ DHCP_print_option $tag ] ($tag)"
         puts "DHCP: Option Length             = $length"
         puts "DHCP: Option Value              = $data"
      }
   }
}

proc DHCP_print_parameter_list { data length } {
    for { set i 0 } { $i < $length } { incr i } {
       set parm [ string range $data [ expr $i * 2 ] [ expr $i * 2 + 1 ]]
       set parmStr [ DHCP_parameter_type $parm ]
       puts "DHCP:   [format "%-23s" $parmStr] = 0x$parm"
    }     
}


proc DHCP_print_option { tag } {
 
   set str "Unknown"

   if { $tag == "55" } {
      set str "Parameter List"
   } elseif { $tag == "60" } {
      set str "Vendor Class Id"
   } elseif { $tag == "81" } {
      set str "Client FQDN"
   }

   return $str
}

proc DHCP_op_type { type } {

   if { $type == "01" } {
      set str "Request"
   } elseif { $type == "02" } {
      set str "Reply"
   } else {
      set str "Unknown DHCP Type ($type)"
   }

   return $str
}

proc DHCP_parameter_type { typehex } {
    set type [ expr 0x$typehex ]

    if { "$type" == "1" } {
      set str "Subnet Mask"
    } elseif { "$type" == "2" } {
      set str "Time Offset"
    } elseif { "$type" == "3" } {
      set str "Router Option"
    } elseif { "$type" == "4" } {
      set str "Time Server Option"
    } elseif { "$type" == "5" } {
      set str "Name Server Option"
    } elseif { "$type" == "6" } {
      set str "Domain Name Server"
    } elseif { "$type" == "12" } {
      set str "Host Name Option"
    } elseif { "$type" == "15" } {
      set str "Domain Name"
    } else {
      set str $type
    }

    return $str
}


proc DHCP_get_message_type { type } {

#
#           Value   Message Type
#           -----   ------------
#             1     DHCPDISCOVER
#             2     DHCPOFFER
#             3     DHCPREQUEST
#             4     DHCPDECLINE
#             5     DHCPACK
#             6     DHCPNAK
#             7     DHCPRELEASE
#

   if { $type == "01" } {
      set str "DHCPDISCOVER"
   } elseif { $type == "02" } {
      set str "DHCPOFFER"
   } elseif { $type == "03" } {
      set str "DHCPREQUEST"
   } elseif { $type == "04" } {
      set str "DHCPDECLINE"
   } elseif { $type == "05" } {
      set str "DHCPACK"
   } elseif { $type == "06" } {
      set str "DHCPNAK"
   } elseif { $type == "07" } {
      set str "DHCPRELEASE"
   } else {
      set str "Unknown DHCP Message Type ($type)"
   }
}

proc DHCP_add_mtype_option { options type } {
   upvar $options o

   set i $o(count) 
   set o(option,$i,tag) [ format "%2.2x" 53 ]
   set o(option,$i,length) 01
   set o(option,$i,data) $type

   incr o(count)
}

#
# DHCP_add_dns_option
#
# Add upto 3 DNS servers
#

proc DHCP_add_dns_option { options dns1 dns2 dns3 } {
   upvar $options o

   set i $o(count)

   set len 0
   set data ""

   if { "$dns1" !=  "0.0.0.0" } {
      incr len 4
      set data "${data}[iptohex $dns1]"
   }

   if { "$dns2" !=  "0.0.0.0" } {
      incr len 4
      set data "${data}[iptohex $dns2]"
   }

   if { "$dns3" !=  "0.0.0.0" } {
      incr len 4
      set data "${data}[iptohex $dns3]"
   }
 
   set o(option,$i,tag) [ format "%2.2x" 06 ]
   set o(option,$i,length) [ format "%2.2x" $len ]
   set o(option,$i,data) $data

   incr o(count)
}

proc DHCP_add_name_option { options name } {
   upvar $options o

   set len [ expr [ string length $name ] + 1]
   set nameHex [ pktsrc text2ascii $name ]

   set i $o(count) 
   set o(option,$i,tag) [ format "%2.2x" 12 ]
   set o(option,$i,length) [ format "%2.2x" $len ]

   # -- null terminated string
   set o(option,$i,data) ${nameHex}00

   incr o(count)
}

proc DHCP_add_option { options tag length data } {

   upvar $options o

   set i $o(count) 
   set o(option,$i,tag) [ format "%2.2x" $tag ]
   set o(option,$i,length) [ format "%2.2x" $length ]
   set o(option,$i,data) $data

   incr o(count)

}


proc DHCP_send_discovery { stack } {
   upvar $stack s

   set xid [ DHCP_get_xid ]

   set header [ DHCP_build_header s 01 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPDISCOVER
   DHCP_add_mtype_option options 01

   # -- add hostname option
   DHCP_add_name_option options $s(DHCP,hostname) 

   DHCP_add_option options 61 7 "01${s(mac)}"

   DHCP_add_option options 55 4 0103060f

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s "Sending DHCPDISCOVER with xid $xid"

   # -- send this in an UDP packet
   DHCP_send s "${header}${opt}" [iptohex 255.255.255.255 ] 67 68

   return $xid
}

proc DHCP_send_offer { stack xid clientMac serverIp { relayIp 0.0.0.0 } } {
   upvar $stack s

   # -- set the clients MAC to be used
   set s(DHCP,server,clientMac) $clientMac

   # -- this is the IP to assign to the client
   set clientIp $s(DHCP,server,address)

   set s(DHCP,yourIP) $clientIp

   # -- build the DHCP header
   set header [ DHCP_build_header s 02 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPOFFER
   DHCP_add_mtype_option options 02

   for { set j 0 } { $j < $s(DHCP,site_option_count) } { incr j } {

       DHCP_add_option options $s(DHCP,site_option,${j},type) $s(DHCP,site_option,${j},length) $s(DHCP,site_option,${j},value) 
   }

   # -- add the mask
   DHCP_add_option options 01 4 [ iptohex $s(DHCP,server,mask) ]

   # -- add the default routers
   DHCP_add_option options 03 4 [ iptohex $s(DHCP,server,gateway) ]

   # -- add dns
   DHCP_add_dns_option options $s(DHCP,server,dnsServer) $s(DHCP,server,dnsServer2) $s(DHCP,server,dnsServer3)

   # DHCP_add_option options 06 4 [ iptohex $s(DHCP,server,dnsServer) ]

   # -- add domain
   set length [ string length $s(DHCP,server,domain) ]
   DHCP_add_option options 15 $length [pktsrc text2ascii $s(DHCP,server,domain) ]

   # -- add lease
   DHCP_add_option options 51 4 [ format "%8.8x" $s(DHCP,server,lease) ]

   # -- add server identifier
   DHCP_add_option options 54 4 [ iptohex $serverIp ]

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s \
      "Sending DHCPOFFER to [hextomac $clientMac] for $clientIp (xid $xid)"

   if { "$relayIp" != "0.0.0.0" } {
      pktsrcInfo s "Sending DHCPOFFER to $relayIp"
      DHCP_send s "${header}${opt}" [ iptohex $relayIp ] 67 67

   } else {
      # -- Add an ARP entry for this IP
      ARP_add_entry s [ iptohex $clientIp ] $clientMac

      # -- send this in a unicast IP/UDP packet
      DHCP_send s "${header}${opt}" [ iptohex $clientIp ] 68 67
   }

   return $xid
}

proc DHCP_send_ack { stack xid clientMac serverIp { relayIp 0.0.0.0 }} {
   upvar $stack s

   # -- set the clients MAC to be used
   set s(DHCP,server,clientMac) $clientMac

   # -- this is the IP to assign to the client
   set clientIp $s(DHCP,server,address)

   set s(DHCP,yourIP) $clientIp

   # -- build the DHCP header
   set header [ DHCP_build_header s 02 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPACK
   DHCP_add_mtype_option options 05

   for { set j 0 } { $j < $s(DHCP,site_option_count) } { incr j } {
      DHCP_add_option options $s(DHCP,site_option,${j},type) $s(DHCP,site_option,${j},length) $s(DHCP,site_option,${j},value) 
   }

   # -- add the mask
   DHCP_add_option options 01 4 [ iptohex $s(DHCP,server,mask) ]

   # -- add the default routers
   DHCP_add_option options 03 4 [ iptohex $s(DHCP,server,gateway) ]

   # -- add dns
   # DHCP_add_option options 06 4 [ iptohex $s(DHCP,server,dnsServer) ]

   DHCP_add_dns_option options $s(DHCP,server,dnsServer) $s(DHCP,server,dnsServer2) $s(DHCP,server,dnsServer3)

   # -- add domain
   set length [ string length $s(DHCP,server,domain) ]
   DHCP_add_option options 15 $length [pktsrc text2ascii $s(DHCP,server,domain) ]

   # -- add lease
   DHCP_add_option options 51 4 [ format "%8.8x" $s(DHCP,server,lease) ]

   # -- add server identifier
   DHCP_add_option options 54 4 [ iptohex $serverIp ]

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s \
      "Sending DHCPACK to [ hextomac $clientMac ] for $clientIp with xid $xid"

   # -- if needed, relay back to dhcp relay server 
   if { "$relayIp" != "0.0.0.0" } {
       pktsrcInfo s "Sending DHCPACK to $relayIp"
       DHCP_send s "${header}${opt}" [ iptohex $relayIp ] 67 67

   } else {

      # -- Add an ARP entry for this IP
      ARP_add_entry s [ iptohex $clientIp ] $clientMac

      # -- send this in a unicast IP/UDP packet
      DHCP_send s "${header}${opt}" [ iptohex $clientIp ] 68 67
   }

   return $xid
}

proc DHCP_send_nack { stack xid clientMac serverIp { relayIp 0.0.0.0 } } {
   upvar $stack s

   # -- build the DHCP header
   set header [ DHCP_build_header s 02 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPNACK
   DHCP_add_mtype_option options 06

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s \
     "Sending DHCPNACK to [ hextomac $clientMac ] with xid $xid"

   if { "$relayIp" != "0.0.0.0" } {
      pktsrcInfo s "Sending DHCPNACK to $relayIp"
      DHCP_send s "${header}${opt}" [ iptohex $relayIp ] 67 67
   } else {
      # -- send this in a broadcast IP/UDP packet
      DHCP_send s "${header}${opt}" [ iptohex 255.255.255.255 ] 68 67
   }
   return $xid
}


proc DHCP_get_xid { } {
   return [ format "%8.8x" [ clock clicks ]]
}

proc DHCP_send_request { stack xid server address } {
   upvar $stack s

   set header [ DHCP_build_header s 01 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPREQUEST
   DHCP_add_mtype_option options 03

   # -- add required options for DHCPREQUEST

   # -- server id
   DHCP_add_option options 54 4 [ iptohex $server ] 

   # -- requested address (if specified)
   if { "$address" != "0.0.0.0" } {
      DHCP_add_option options 50 4 [ iptohex $address ]
   }

   
   # DHCP_add_name_option options fred

   # -- use the hostname for this dhcp clients name since it should
   # be unique. By default, the hostname is set to the stack name
   # when during DHCP_init

   DHCP_add_name_option options $s(DHCP,hostname)

   DHCP_add_option options 61 7 "01${s(mac)}"

   DHCP_add_option options 55 4 0103060f

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s "Sending DHCPREQUEST with xid $xid"

   # -- send this in an UDP packet
   DHCP_send s "${header}${opt}" [iptohex 255.255.255.255 ] 67 68

   return $xid
}

proc DHCP_send_release { stack xid server address } {
   upvar $stack s

   set header [ DHCP_build_header s 01 $xid ]

   # -- add my options
   set options(count) 0

   # -- add the message type for DHCPRELEASE
   DHCP_add_mtype_option options 07

   # -- add required options for DHCPRELEASE

   # -- server id
   DHCP_add_option options 54 4 [ iptohex $server ] 

   # -- requested address (if specified)
   if { "$address" != "0.0.0.0" } {
      DHCP_add_option options 50 4 [ iptohex $address ]
   }

   DHCP_add_name_option options $s(DHCP,hostname)

   DHCP_add_option options 61 7 "01${s(mac)}"

   # -- build the options
   set opt [ DHCP_build_options options ]

   pktsrcInfo s "Sending DHCPRELEASE of $address from $server"

   # -- send this in an UDP packet
   DHCP_send s "${header}${opt}" [iptohex $server ] 67 68

   return $xid
}

proc DHCP_send { stack packet iphex sport dport } {
   upvar $stack s

   if { "$s(rfc1541MinimumSize)" == "on" } {
      UDP_send_pad s $packet $iphex $sport $dport 556
   } else {
      UDP_send s $packet $iphex $sport $dport
   }

}

proc DHCP_build_header { stack type xid } {

   upvar $stack s

   set op      $type
   set htype   01
   set hlen    06
   set hops    00
   set xid2    $xid
   set sec     0000

   # -- check for the broadcast flag
   if { "$s(DHCP,broadcast-bit)" == "1" } {
      set flags   8000
   } else {
      set flags   0000
   }

   set ciaddr  00000000
   set yiaddr  00000000
   set siaddr  00000000
   set giaddr  00000000

   # -- fill in the ciaddr if we have one
   if { "$s(DHCP,clientIP)" != "0.0.0.0" } {
       set ciaddr [ iptohex $s(DHCP,clientIP) ]
   }

   # -- fill in the giaddr if we have one
   if { "$s(DHCP,relayIp)" != "0.0.0.0" } {
       set giaddr [ iptohex $s(DHCP,relayIp) ]
   }

   # -- fill in the yiaddr if we have one
   if { "$s(DHCP,yourIP)" != "0.0.0.0" } {
       set yiaddr [ iptohex $s(DHCP,yourIP) ]
   }


   if { "$s(DHCP,server,clientMac)" == "000000000000" } {
     set chaddr  $s(mac)
   } else {
      set chaddr $s(DHCP,server,clientMac)
   }

   # set sname
   # set file 

   set notused [ string repeat 00 [ expr 10 + 64 + 128 ]]

   return ${op}${htype}${hlen}${hops}${xid2}${sec}${flags}${ciaddr}${yiaddr}${siaddr}${giaddr}${chaddr}${notused}
 
}


proc DHCP_build_options { op } {
   upvar $op o

   set cookie [ iptohex 99.130.83.99 ]
   set oplist $cookie

   for { set i 0 } { $i < $o(count) } { incr i } {

       set tag    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)      
       set oplist ${oplist}${tag}${length}${data}      

   }     

   # -- add an end marker -- ff
   set oplist "${oplist}ff"

   # -- see if we need to pad option list to be even boundary
   set len [ string length $oplist ]
   if { [ expr ($len/2) % 2 ] == 1 } {
       set oplist "${oplist}00"
   }

   return $oplist
}

#
# DHCP_client
#
# Start the DHCP client
#
#   if timeOut > 0, block until DHCPACK or timeOut fires
#   otherwise, return immediately
#

proc DHCP_client { stack { timeOut 0 } { hostname "" } } {
   upvar $stack s

   pktsrcInfo s "Starting DHCP client"

   # -- remember the broadcast bit setting, this is sticky
   set bb $s(DHCP,broadcast-bit)

   DHCP_init s

   # -- reset the broadcast bit
   set s(DHCP,broadcast-bit) $bb

   # -- set the hostname if specified
   if { "$hostname" != "" } {
      set s(DHCP,hostname) $hostname
   }

   if [ info exist s(DHCP,handler) ] {
      # -- handler already installed
   } else {
      Stack_add_callback s dhcpType 02 DHCP_reply_input
      set s(DHCP,handler) 1
   }

   #
   # -- block if a timeout was requested

   set totalTime 0
   if { "$timeOut" > 0 } {

      pktsrcInfo s "Waiting upto $timeOut msec to obtain IP address"

      # -- wait for a DHCPACK or timeout
      
      while { $totalTime < $timeOut } {
         set xid [ DHCP_send_discovery s ]
         set s(DHCP,xid) "$xid"
         # -- store xids in hex for direct comparison
         set s(DHCP,client,xid,discovery,$xid) $xid

         set result [ DHCP_get_packet s 05 dhcp 5 ]

	 if { $result != "-1" } {
            break
	 }
         incr totalTime 5000
      }

      # -- process any remaining events
      Event_wait 1000

   } else {

     # -- otherwise, just start the discovery

     set xid [ DHCP_send_discovery s ]
     set s(DHCP,xid) "$xid"
     # -- store xids in hex for direct comparison
     set s(DHCP,client,xid,discovery,$xid) $xid
   }

   return $s(DHCP,clientIP)
}
    

proc DHCP_renew { stack } {   
   upvar $stack s

   # -- get xid (same xid is reused)
   set xid $s(DHCP,xid)

   pktsrcInfo s "Attempting to renew lease for $s(DHCP,clientIP) from $s(DHCP,serverIP)"

   DHCP_send_request s $xid $s(DHCP,serverIP) $s(DHCP,clientIP)

   set result [ DHCP_get_packet s 05 dhcp 15 ]

   if { "$result" != "-1" } {
      pktsrcInfo s "lease for client $s(DHCP,clientIP) renewed by server $s(DHCP,serverIP)"
      return 0
   } else {
      pktsrcInfo s "Failed to renew lease on $s(DHCP,clientIP)"
      return -1
   }
}


proc DHCP_client_release { stack } {
   upvar $stack s

   if { $s(DHCP,clientIP) == "0.0.0.0" } {
       pktsrcWarn s "No client IP to release"
       return
   } 

   set xid $s(DHCP,xid)
   set server $s(DHCP,serverIP)
   set addr $s(DHCP,clientIP)

   pktsrcInfo s "Releasing lease for $s(DHCP,clientIP) from $s(DHCP,serverIP)"
   DHCP_send_release s $xid $server $addr

   DHCP_init s
}

proc DHCP_broadcast_mode { stack { mode on } } {
   upvar $stack s

   if { "$mode" == "on" } {
       set s(DHCP,broadcast-bit) 1
   } else {
       set s(DHCP,broadcast-bit) 0
   }
}

proc DHCP_find_option { dhcp tag } {
   upvar $dhcp o 
   set result ""

   for { set i 0 } { $i < $o(options) } { incr i } {
       set t    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)             

	if { "[expr 0x$t]" == "$tag" } {
          set result $data
          break
        }
   }

   return $result
}

proc DHCP_find_dns_option { dhcp dns1 dns2 dns3 } {
   upvar $dhcp o 

   upvar $dns1 d1
   upvar $dns2 d2
   upvar $dns3 d3

   
   set result [ DHCP_find_option o 06 ] 
   set len [ string length $result ]

   if { $len > 0 } {

      if { $len >= 8 } {
         set d1 [hextoip [ string range $result 0 7 ]]
      } else {
         set d1 0.0.0.0
      }

      if { $len >= 16 } {
         set d2 [hextoip [ string range $result 8 15 ]]
      } else {
         set d2 0.0.0.0
      }

      if { $len >= 24 } {
         set d3 [hextoip [ string range $result 16 23 ]]
      } else {
         set d3 0.0.0.0
      }
   }

   return $result
}

#
# DHCP_server
#
# enable a simple DHCP server service on this stack
# only a single DHCP client is support
# the client will be assigned the 'address'
#

proc DHCP_server { stack address mask seconds domain dnsServer gateway } {
   upvar $stack s

   pktsrcInfo s "Starting DHCP server"

   DHCP_init s

   set s(DHCP,server,mode) on

   #
   # -- remember the server configuration

   set s(DHCP,server,address) $address
   set s(DHCP,server,mask) $mask
   set s(DHCP,server,lease) $seconds
   set s(DHCP,server,domain) $domain
   set s(DHCP,server,dnsServer) $dnsServer
   set s(DHCP,server,dnsServer2) 0.0.0.0
   set s(DHCP,server,dnsServer3) 0.0.0.0
   set s(DHCP,server,gateway) $gateway

   #
   # -- install a handler for dhcp requests

   if [ info exist s(DHCP,serverHandler) ] {

      # -- handler already installed

   } else {

      Stack_add_callback s dhcpType 01 DHCP_request_input
      set s(DHCP,serverHandler) 1
   }

}

proc DHCP_add_site_option { stack type length value } {
   upvar $stack s

   set j $s(DHCP,site_option_count)

   set s(DHCP,site_option,$j,type) $type 
   set s(DHCP,site_option,$j,length) $length
   set s(DHCP,site_option,$j,value) $value

   incr s(DHCP,site_option_count)
}

proc DHCP_remove_site_options { stack } {
   upvar $stack s

   set s(DHCP,site_option_count) 0
}

proc DHCP_server_set_dns { stack dns1 dns2 dns3 } {
   upvar $stack s

   set s(DHCP,server,dnsServer)  $dns1
   set s(DHCP,server,dnsServer2) $dns2
   set s(DHCP,server,dnsServer3) $dns3

}

proc DHCP_server_disable { stack } {
   upvar $stack s

   pktsrcInfo s "Disable DHCP server"
   set s(DHCP,server,mode) off

}

proc DHCP_server_enable { stack } {
   upvar $stack s

   pktsrcInfo s "Enable DHCP server"
   set s(DHCP,server,mode) on

}

#
# DHCP_Server_lease_time
#
# return the time remaining on a lease for the DHCP client
# NOTE: the DHCP server only supports one client, so no client identifier 
# is needed.
#

proc DHCP_server_lease_time { stack } {
   upvar $stack s

   set ct [ clock seconds ]
   set timeLeft [ expr $s(DHCP,server,leaseEnd) - $ct ]

   if { "$timeLeft" < "0" } {
      return 0
   } else {
      return $timeLeft
   }
   
}

#
# DHCP_server_wait_for_renew
#
# Wait for the current client to renew its lease
#
# This can be used to wait for a new lease or to renumber the client
#

proc DHCP_server_wait_for_renew { stack { verifyStack "" }} {
    upvar $stack s

    if { "$verifyStack" != "" } {
      upvar $verifyStack lan
    } else {
      set lan ""
    }

    pktsrcInfo s "Waiting for DHCP client to renew lease"

    set leaseTime [ DHCP_server_lease_time s ] 

    pktsrcInfo s "The current lease has $leaseTime seconds remaining"

    # -- if time remaining, wait for it to expire
    if { "$leaseTime" > "0" } {

       pktsrcInfo s "waiting $leaseTime seconds for lease to expire"

       # -- event wait is msec
       Event_wait [ expr $leaseTime * 1000 ]
    }

    # -- verify the client address is active
    if { "$verifyStack" != "" } {
       for { set c 0 } { $c < 18 } { incr c } {
           set address [ ICMP_ping lan $s(DHCP,server,address) ]
	   if { "$address" != "0" } {
              break
           } else {
              Event_wait 5000
           }
       }
    } else {

       for { set c 0 } { $c < 18 } { incr c } {
           set address [ ARP_get_wait s $s(DHCP,server,address) ]
	   if { "$address" != "" } {
              break
           } else {
              Event_wait 5000
           }
       }
    }
   
    if { "$address" == "" || "$address" == "0" } {
       pktsrcWarn s "DHCP client address IP address is not reachable"
    } else {
       pktsrcInfo s "DHCP client address is now reachable"
    }
}


#
# DHCP helper procs
#

#
# DHCP_generic_input
#
# generic callback for catching DHCP packets
#

proc DHCP_generic_input { packet stack { options none }} {
   upvar $stack s

   if { "$options" != "print" } {

      # -- trigger the packet event
      Event_got_packet
   }
}

#
# DHCP_get_packet
#
# blocking call for receiving DHCP packets
# this call blocks until the packet is received or a timeout occurs
#

proc DHCP_get_packet { stack type packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for DHCP '$type' packets
   Stack_add_callback s dhcp53Type $type DHCP_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout seconds for [DHCP_get_message_type $type ] packet"

   # -- compute timeout in milli seconds
   set msec [ expr $timeout * 1000 ]

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the DHCP packet from the stack
      set dhcpState $s(current,DHCP)
      upvar #$s(current,DHCP,level) $dhcpState dhcp
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet dhcp p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "$timeout second timeout waiting for [ DHCP_get_message_type $type ] packet"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s dhcp53Type $type DHCP_generic_input

   return $ret
}








