#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: eth.tcl,v 1.13 2002/11/15 13:23:54 joe Exp $
#
#
# Basic Ethernet II
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3


proc Ethernet_input { packet stack timeSec timeUsec } {

   upvar $stack s

   # -- remember this packet in the 'eth' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s Ethernet eth ]

   #set s(current,Ethernet) eth
   #set s(current,Ethernet,level) [info level]
   
   # -- decode ethernet  
   Ethernet_decode $packet eth

   # -- set the current timestamps for this packet
   set s(current,timestampSec) $timeSec
   set s(current,timestampUsec) $timeUsec

   # -- get the port
   set port $s(port)

   # -- need to make printing option
   if [ info exists s(print) ] {
       puts "\nPACKET-IN([$port device]): received eth proto type $eth(type)"
       Ethernet_print eth
       flush stdout
   }


   if [ info exists s(ethType,$eth(type)) ] {

      if [ info exists s(print) ] {
         # -- drive any printing callbacks
         if [ catch { $s(ethType,$eth(type),input) \
                             $eth(data) s print } reason ] {
           puts "PKTSRC-ERROR: print callback '$s(ethType,$eth(type),input)' failed"
           puts "PKTSRC-ERROR: $reason"
         }

         puts ""     
      }
      
      # -- drive the real input callbacks as long as the stack is
      # not in a sniffer mode

      if ![ info exists s(snifferMode) ] {
      # -- drive any normal input callbacks
        foreach c $s(ethType,$eth(type),callbacks) {


         if [ catch { $c $eth(data) s } reason ] {
            puts "PKTSRC-ERROR: input callback '$s(ethType,$eth(type),input)' failed"
            puts "PKTSRC-ERROR: $reason"
         }
        }
      }

      # -- more generic callbacks
   } else {
      # no callback
   }

   # -- if packet trace

}

proc Ethernet_output { packet type dst stack } {

   upvar $stack s

   # -- get the port
   set port $s(port)

 
   # -- build an Ethernet packet
   # set out [ format "%s%s%s%s" ]
   
   if { $type == "0806" } {
      # -- this is ARP
      set out [ format "%s%s%s%s" ] $dst $s(mac) $type $packet
      $port send $out
   } else {
      puts "Ethernet_output: can't send unknown type"
   }

}

proc Ethernet_log_send { packet stack } {

   upvar $stack s 

   # -- only if the print option is enabled

   if [ info exists s(print) ] {
      # -- decode ethernet  
      Ethernet_decode $packet eth

      # -- get the port
      set port $s(port)

      puts "\nPACKET-OUT([$port device]): sending proto type $eth(type)"
      Ethernet_print eth

      if [ info exists s(ethType,$eth(type)) ] {

         # -- drive any printing callbacks
         if [ catch { $s(ethType,$eth(type),input) $eth(data) s print } \
           reason ] {
           puts "PKTSRC-ERROR: log send callback '$s(ethType,$eth(type),input)' failed"
           puts "PKTSRC-ERROR: $reason"
         }
      }
      puts ""
   }
}


proc Ethernet_decode { packet eth } {
   upvar $eth out

   set out(totalLen)  [ expr [ string length $packet ] / 2 ]
   # -- parse out all the fields we find interesting ...

   set out(dstMac)    [ string range $packet 0 11 ]
   set out(srcMac)    [ string range $packet 12 23 ]
   set out(type)      [ string range $packet 24 27 ]

   # -- check for vlan tag
   if { "$out(type)" == "8100" } {

      set out(vlanType)  $out(type)
      set out(vlanData)  [ string range $packet 28 31 ]
      set out(type)      [ string range $packet 32 35 ]
      set out(data)      [ string range $packet 36 end ]    

      puts "new type is $out(type)"      
   } else {

      set out(vlanType)  0000
      set out(vlanData)  0000
      set out(data)      [ string range $packet 28 end ]    
   }


}

proc Ethernet_print { eth } {
   upvar $eth out

   puts "ETH: ---- Ethernet Packet (length = $out(totalLen)) ----"
   puts "ETH: "
   puts "ETH: Source                     = [hextomac $out(srcMac)]"
   puts "ETH: Destination                = [hextomac $out(dstMac)]"


   if { "$out(vlanType)" == "8100" } {
    puts "ETH: 802.1q Tag                 = 0x$out(vlanType)"
    puts "ETH: 802.1q Priority            = [expr 0x$out(vlanData) >> 13 ]"
    puts "ETH: 802.1q CFI                 = [expr (0x$out(vlanData) & 0x1ff) >> 12]"
    puts "ETH: 802.1q VID                 = [expr 0x$out(vlanData) & 0x0fff]"
   }

   puts "ETH: Type                       = [Ethernet_protoType $out(type)] (0x$out(type))"
   puts "ETH:"


}

proc Ethernet_protoType { type } {

   if { $type == "0800" } {
      set str "IPv4"
   } elseif { $type == "0806" } {
      set str "ARP"
   } elseif { $type == "8100" } {
      set str "802.1q"
   } elseif { $type == "8806" } {
      set str "IPv6"
   } elseif { $type == "8863" } {
      set str "PPPoE Discovery"
   } elseif { $type == "8864" } {
      set str "PPPoE Session"
   } elseif { [ expr 0x$type ] < 1500 } {
      set str "802.3 Ethernet"
   } else {
      set str "Unknown Protocol ($type)"
   }
   return $str
}

proc Ethernet_IPv4_send { packet stack dest } {
   upvar $stack s

   set port $s(port)

   # -- do I have a MAC address for dest
   set mac [ ARP_get s $dest ]

   # -- routing function ?

   if { "$mac" != "" } {
       set epack [ format "%s%s%s%s" $mac $s(mac) "0800" $packet ]
       $port send $epack
       Ethernet_log_send $epack s
   } else {

       # -- if there is already a packet waiting, then the waiting
       # packet is going to get dropped

       if [ info exists s(arpBuffer) ] {
          pktsrcWarn s "Dropping packet to [hextoip $s(arpBuffer,dest)], no MAC in cache"
       }

       # -- buffer this packet       
       set s(arpBuffer) $packet
       set s(arpBuffer,dest) $dest

   }
      
}

proc Ethernet_send { stack packet type macdest } {
   upvar $stack s

   set port $s(port)

   # -- check for vlan tags
   if [ info exists s(Ethernet,802.1q) ] {

      set epack [ format "%s%s%s%s%s" $macdest $s(mac) \
          $s(Ethernet,802.1q) $type $packet ]

   } else {

      set epack [ format "%s%s%s%s" $macdest $s(mac) $type $packet ]
   }

   $port send $epack

   Ethernet_log_send $epack s
   
}

#
# set a 802.1q vlan for an ethernet interface
#
# tag should be the full 2 byte tag value in hex
#

proc Ethernet_set_8021q_vlan { stack tag } {
   upvar $stack s

   set s(Ethernet,802.1q) "8100${tag}"
}













