#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: http.tcl,v 1.8 2002/11/15 13:23:54 joe Exp $
#
#
# Poor man's Http
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc HTTP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      HTTP_decode $packet http
      HTTP_print http
      return

   } else {

      # -- remember this packet in the 'http' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s HTTP http ]
   
      # -- now decode the packet into the 'http' variable
      HTTP_decode $packet http

   }

}


#
#
# HTTP_init - set up any deaults for HTTP
#

proc HTTP_init { stack } {
   upvar $stack s

}

proc HTTP_decode { packet http } {
   upvar $http out

   # -- just treat everything as one text string for now

   set out(data)        [ string range $packet 0 end ]

}

proc HTTP_print { http } {
   upvar $http out

   if { $out(data) != "" } {

     puts "HTTP: ---- HTTP Packet ----"
     puts "HTTP: "

     # -- display everything as one text string
     puts "[pktsrc ascii2text $out(data) ]"

     puts "HTTP: "
     puts "HTTP: ---- End of HTTP ----"

   }
}

proc HTTP_send { stack tcpSession data } {
   upvar $stack s
   upvar $tcpSession session

   set hex [ pktsrc text2ascii $data ]

   # -- add 2 carriage return and line feeds at the end of the data
   return [ TCP_write s session "${hex}0d0a0d0a" ]

}

proc HTTP_get_url { stack tcpSession host url { version 1.1 } { cookie "" }} {

   upvar $stack s
   upvar $tcpSession session

   set command(1) "GET $url HTTP/$version"
   set command(2) "Connection: Keep-Alive"
   set command(3) "User-Agent: lbrouter"
   set command(4) "Host: $host"
   set command(5) "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, */*"
   set command(6) "Accept-Encoding: gzip"
   set command(7) "Accept-Language: en"
   set command(8) "Accept-Charset: iso-8859-1,*,utf-8"

   set lines 8
   if { "$cookie" != "" } {
      incr lines
      set command($lines) "Cookie: $cookie"
   }

   set commandhex ""

   for { set i 1 } { $i <= $lines } { incr i } {
       set hex [ pktsrc text2ascii $command($i) ]
       set commandhex "${commandhex}${hex}0d0a"
   }

   # -- add trailing carriage return
   set commandhex "${commandhex}0d0a"

   return [ TCP_write s session "$commandhex" ]
}

proc HTTP_head_url { stack tcpSession host url { version 1.1 } } {

   upvar $stack s
   upvar $tcpSession session

   set command(1) "HEAD $url HTTP/$version"
   set command(2) "Connection: Keep-Alive"
   set command(3) "User-Agent: lbrouter"
   set command(4) "Host: $host"
   set command(5) "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, */*"
   set command(6) "Accept-Encoding: gzip"
   set command(7) "Accept-Language: en"
   set command(8) "Accept-Charset: iso-8859-1,*,utf-8"

   set commandhex ""

   for { set i 1 } { $i <= 8 } { incr i } {
       set hex [ pktsrc text2ascii $command($i) ]
       set commandhex "${commandhex}${hex}0d0a"
   }

   # -- add trailing carriage return
   set commandhex "${commandhex}0d0a"

   return [ TCP_write s session "$commandhex" ]
}

proc HTTP_post_url { stack tcpSession host url { version 1.1 } } {

   upvar $stack s
   upvar $tcpSession session

   set command(1) "POST $url HTTP/$version"
   set command(2) "Connection: Keep-Alive"
   set command(3) "User-Agent: lbrouter"
   set command(4) "Host: $host"
   set command(5) "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, */*"
   set command(6) "Accept-Encoding: gzip"
   set command(7) "Accept-Language: en"
   set command(8) "Accept-Charset: iso-8859-1,*,utf-8"

   set commandhex ""

   for { set i 1 } { $i <= 8 } { incr i } {
       set hex [ pktsrc text2ascii $command($i) ]
       set commandhex "${commandhex}${hex}0d0a"
   }

   # -- add trailing carriage return
   set commandhex "${commandhex}0d0a"

   return [ TCP_write s session "$commandhex" ]
}

#
# HTTP_read
#
# read data from an http connection
# timeout after 'timeOut' if no data is received
#

proc HTTP_read { stack tcpSession data { timeOut 30000 } } {
   upvar $stack s
   upvar $tcpSession session
   upvar $data dataText

   set bytes 0
   set dataBuf ""
   set dataText ""

   #
   # While the session is open, keep reading data
   # No Parsing is implemented

   set elapsed 0

   while { $elapsed < $timeOut } {

      set dataIn [ TCP_get_data s session ]
      set dataBuf "$dataBuf$dataIn"

      # -- check the state
      set state [ TCP_state s session ]

      # -- keep reading while we are establsihed
      if { "$state" != "established" } {
         break
      }
 
      # -- run the event loop (blocking call)
      Event_wait 100
      incr elapsed 100

   }

   # -- convert data to text
   set dataText [ pktsrc ascii2text $dataBuf ]
   set bytes [ string length $dataText ]

   if { $elapsed >= $timeOut } {
     pktsrcWarn s "timeout waiting for data in HTTP_read"
   }

   # -- return the number of bytes
   return $bytes
}


#
# HTTP_dummy_server
#
# Start a dummy HTTP server on the specified port
#
# 

proc HTTP_dummy_server { stack { port 80 }} {
   upvar $stack s

   set s(HTTP,dummyServer) 1

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Starting up dummy web server on port $port"

   TCP_server s $port HTTP_server_input

}

proc HTTP_dummy_server_shutdown { stack { port 80 }} {
   upvar $stack s

   set s(HTTP,dummyServer) 0

   set srcPort [ format "%4.4x" $port ]

   pktsrcInfo s "Shuting down dummy web server on port $port"

   TCP_server_shutdown s $port HTTP_server_input

}

#
# HTTP_server_input
#
# This is the input routine for an http session
# Don't bother parsing the input request, just return the dummy web page
#

proc HTTP_server_input { packet stack stack {options none}} {
   upvar $stack s

   if { "$options" == "print" } {
       return
   }

   set httpState $s(current,HTTP)
   upvar #$s(current,HTTP,level) $httpState http

   if { $http(data) != "" } {

      # -- dummy server is running, process request

      # -- get the current IP contents
      set ipState $s(current,IPv4)
      upvar #$s(current,IPv4,level) $ipState ip

      # -- get the current tcp contents
      set tcpState $s(current,TCP)
      upvar #$s(current,TCP,level) $tcpState tcp

      # -- check that this is a valid session
      set sid \
        [ TCP_get_session s session $ip(src) $tcp(srcPort) $tcp(dstPort) ]

      if { "$sid" != "0" } {

        # -- send a simple web page
        set page "<html><title>Web server at [hextoip $s(ip)]</title><b>Hello</b></br>"
	set page "${page}client ip address: [hextoip $ip(src)]<br>"
	set page "${page}source port: [expr 0x$tcp(srcPort)]<br>"
	set page "${page}page generated: [clock format [clock seconds]]<br>"
	set page "${page}my ip address is [hextoip $s(ip)]<br>"
	set page "${page}</html>"
        
        HTTP_send s session "$page"

        # -- need to fix fin problem
        # Event_wait 500

        # -- close the session
        TCP_close s session

      }
   }
}

#
# URL_get_domain
#
# return the domain portion of a URL
#

proc URL_get_domain { url } {

   set parts [ split $url / ]
   
   set count [ llength $parts ]

   if { "$count" == "1" } {
     return $url
   } elseif { "$count" == "2" } {
     return [ lindex $parts 0 ]
   } else {
     return [ lindex $parts 2 ]
   }

}

#
# URL_get_page
#
# return the page portion of a URL
#

proc URL_get_page { url } {

   set parts [ split $url / ]
   
   set count [ llength $parts ]

   if { "$count" == "1" } {

       return "/"

   } elseif { "$count" == "2" } {

       return "/[lindex $parts 1]"

   } elseif { "$count" == "3" } {

       return "/"

   } else {

       set page ""
       for { set i 3 } { $i < $count } { incr i } {
	  set page "${page}/[lindex $parts $i]"
       }

       return $page

   }
}





