#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: ipsec.tcl,v 1.4 2002/11/15 13:23:54 joe Exp $
#
#
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc IKE_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      if [ catch { IKE_decode $packet ike } reason ] {
         pktsrcWarn s "Unable to fully parse IKE packet"
      } else {
         IKE_print ike
      }
      return

   } else {

      # -- remember this packet in the 'ike' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s IKE ike ]
   
      # -- now decode the packet into the 'ike' variable
      if [ catch { IKE_decode $packet ike } reason ] {
         pktsrcWarn s "Unable to fully parse IKE packet"
         return
      }

   }

   # -- check for callbacks at the IKE level
   # if [ info exists s(ikeType,$ike(command)) ] {
   #
   #   # -- drive any input callbacks
   #   foreach c $s(ikeType,$ike(command),callbacks) {
   #       # -- send the whole packet
   #       $c $packet s $options
   #   }
   #
   # } else {
   #    # no callback
   # }
}

proc IKE_init { stack } {
   upvar $stack s

}


#
#                         1                   2                   3
#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    !                          Initiator                            !
#    !                            Cookie                             !
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    !                          Responder                            !
#    !                            Cookie                             !
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    !  Next Payload ! MjVer ! MnVer ! Exchange Type !     Flags     !
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    !                          Message ID                           !
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    !                            Length                             !
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#
#                 Figure 2:  ISAKMP Header Format
#
#    o  Next Payload (1 octet) - Indicates the type of the first payload
#       in the message.  The format for each payload is defined in
#       sections 3.4 through 3.16.  The processing for the payloads is
#       defined in section 5.
#
#
#                        Next Payload Type       Value
#                    NONE                           0
#                    Security Association (SA)      1
#                    Proposal (P)                   2
#                    Transform (T)                  3
#                    Key Exchange (KE)              4
#                    Identification (ID)            5
#                    Certificate (CERT)             6
#                    Certificate Request (CR)       7
#                    Hash (HASH)                    8
#                    Signature (SIG)                9
#                    Nonce (NONCE)                 10
#                    Notification (N)              11
#                    Delete (D)                    12
#                    Vendor ID (VID)               13
#                    RESERVED                   14 - 127
#                    Private USE               128 - 255
#

proc IKE_decode { packet IKE } {
   upvar $IKE out

   # -- parse out all the fields we find interesting ...

   set out(init-cookie)           [ string range $packet 0 15 ]
   set out(respond-cookie)        [ string range $packet 16 31 ]

   set out(next-payload)          [ string range $packet 32 33 ]
   set out(major-ver)             [ string range $packet 34 34 ]
   set out(minor-ver)             [ string range $packet 35 35 ]
   set out(type)                  [ string range $packet 36 37 ]
   set out(flags)                 [ string range $packet 38 39 ]
   set out(id)                    [ string range $packet 40 47 ]
   set out(length)                [ string range $packet 48 55 ]

   # -- check encrypted flag

   set out(nh)                    0
  
   if { [ expr 0x$out(flags) & 0x01 ] } {

     # -- payload is encrypted

   } elseif { "$out(next-payload)" != "00" } {
      IKE_decode_sub out [ string range $packet 56 end ] $out(next-payload)
   }
}

proc IKE_decode_sub { ike packet type } {
   upvar $ike out

   set nh [ string range $packet 0 1 ]
   set out(sub,$out(nh),nh) $nh
   set out(sub,$out(nh),ch) $type

   set length [ expr 0x[ string range $packet 4 7 ]]
   set out(sub,$out(nh),length) $length

   # -- check length
   set myend [ expr $length * 2 - 1 ]

   # puts "LENGTH = $out(sub,$out(nh),length)"
   # puts "MYEND = $myend"

   set out(sub,$out(nh),data) [ string range $packet 0 $myend ]

   if { "$out(sub,$out(nh),nh)" != "00" && $length > 0 } {
      incr out(nh)
      set restOfpacket [ string range $packet [ expr $myend + 1 ] end ]

       # -- more data and next-header
       if { "$restOfpacket" != "" } {
         IKE_decode_sub out $restOfpacket $nh
       }
   } else {
      incr out(nh)
   }
}

proc IKE_print_sa_payload { packet } {

   set doi [ expr 0x[ string range $packet 8 15 ]]
   set sit [ expr 0x[ string range $packet 16 23 ]]

   if { "$doi" != "1" } {

     puts "IKE:    Domain of Interp        = $doi"
     puts "IKE:    Situation               = $sit"

   } else {

     puts "IKE:    Domain of Interp        = ipsec ($doi)"
     puts "IKE:    Situation               = [ IKE_get_sa_sit $sit ] ($sit)"

     IKE_print_proposal_payload [ string range $packet 24 end ]

   }
}

proc IKE_print_vendor_id { packet } {
     
   set vid [ string range $packet 8 end ]
   set vidStr [ pktsrc ascii2text $vid ]
   puts "IKE:    ID Str                  = $vidStr"
}

proc IKE_print_proposal_payload { packet } {

   set number [ expr 0x[ string range $packet 8 9 ]]
   set protocol [ expr 0x[ string range $packet 10 11 ]]
   set spiSize [ expr 0x[ string range $packet 12 13 ]]
   set transforms [ expr 0x[ string range $packet 14 15 ]]


   puts "IKE:    Proposal Number         = $number"
   puts "IKE:    protocol                = $protocol"
   puts "IKE:    SPI size                = $spiSize"
   puts "IKE:    Transforms              = $transforms"


   if { "$spiSize" > "0" } {
      set spi [ string range $packet 16 [ expr 15 + $spiSize * 2  ]]
      puts "IKE:    SPI                     = $spi"
   }

   set myend [ expr 16 + $spiSize * 2 ]

   IKE_print_transform_payload [ string range $packet $myend end ]
}

proc IKE_print_transform_payload { packet } {

   set number [ expr 0x[ string range $packet 8 9 ]]
   set id [ expr 0x[ string range $packet 10 11 ]]

   puts "IKE:    Transform Number        = $number"
   puts "IKE:    ID                      = $id"

   set length [ string length $packet ]
   set start 16

   while { $start < $length } {
     set myend [ expr $start + 3 ]
     set attr [ string range $packet $start $myend ]

     
     incr start 4
     incr myend 4
     set value [ string range $packet $start $myend ]

     if { [ expr 0x8000 & 0x$attr ] } {
 
        # -- no problem
        incr start 4
        set attr [ expr 0x7fff & 0x$attr ]

     } else {
        # -- value is really length

        incr start 4
        incr myend [ expr 0x$value * 2 ]
        set value [ string range $packet $start $myend ]
        set start $myend
        incr start

        set attr [ expr 0x$attr ]
      }

      puts "IKE:    [ IKE_print_isakmp_attr $attr $value]"
      
   }
}

proc IKE_print_isakmp_attr { type value } {

    # -- move this to gloal scope

    set map(1) encryption
    set map(1,1) 1des
    set map(1,2) idea
    set map(1,3) blowfish
    set map(1,4) rc5
    set map(1,5) 3des
    set map(1,6) cast
    set map(1,7) aes

    set map(2) hash
    set map(2,1) md5 
    set map(2,2) sha1 
    set map(2,3) tiger
    set map(2,4) sha2-256
    set map(2,5) sha2-384
    set map(2,6) sha2-512

    set map(3) auth
    set map(3,1) preshared
    set map(3,2) dss
    set map(3,3) rsa-sig
    set map(3,4) rsa-enc
    set map(3,5) rsa-enc-revised

    set map(4) group-desc
    set map(4,1) modp768
    set map(4,2) modp1024
    set map(4,3) "EC2N 2^155"
    set map(4,4) "EC2N 2^185"

    set map(5) group-type
    set map(5,1) MODP
    set map(5,2) ECP
    set map(5,3) EC2N

    set map(11) lifetype
    set map(11,1) seconds
    set map(11,2) kb

    set map(12) lifetime

    set t $type
    set v [ expr 0x$value ]

    if [ info exists map($t,$v) ] {
       return "[ format "%-23s" $map($t)] = $map($t,$v)"
    } elseif [ info exists map($t) ] {
       return "[ format "%-23s" $map($t)] = $value"
    } else {
       return "attr [ format "%-4s" $type ]               = $value"
    }

}

proc IKE_get_sa_sit { sit } {
   set sitStr ""

    if { [ expr $sit & 0x01 ] } {
      set sitStr "identity"
   }

   if { [ expr $sit & 0x02 ] } {
       if { "$sitStr" != "" } {
         set sitStr "${sitStr}+secrecy"
       } else {
         set sitStr "secrecy"
       }
   }

   if { [ expr $sit & 0x04 ] } {
       if { "$sitStr" != "" } {
         set sitStr "${sitStr}+integrity"
       } else {
         set sitStr "integrity"
       }
   }

   return $sitStr

}

proc IKE_print { IKE } {
   upvar $IKE out

   puts "IKE: ---- IKE Packet ----"
   puts "IKE: "
   puts "IKE: Initiator Cookie           = 0x$out(init-cookie)"
   puts "IKE: Responder Cookie           = 0x$out(respond-cookie)"
   puts "IKE: Next Payload               = [ ISAKMP_payload_type $out(next-payload) ] (0x$out(next-payload))"

   puts "IKE: Version                    = [expr 0x$out(major-ver)].[expr 0x$out(minor-ver)]"
   
   if { "$out(id)" == "00000000" } {
      puts "IKE: Phase                      = 1"
   } else {
      puts "IKE: Phase                      = [ expr 0x$out(id)]/other"
   }

   puts "IKE: Flags                      = 0x$out(flags)"

   if { [ expr 0x$out(flags) & 0x01 ] } {
       puts "IKE: Encrypted Payload"
   } else {

      for { set i 0 } { $i < $out(nh) } { incr i } {
         IKE_print_sub out $i
      }
   }

   puts "IKE:"
   puts "IKE: ---- End IKE Packet ----"
}

proc IKE_print_sub { ike pos } {
   upvar $ike out


   puts "IKE: Payload Type               = [ ISAKMP_payload_type $out(sub,$pos,ch) ] (0x$out(sub,$pos,ch))"
   puts "IKE: Payload Length             = $out(sub,$pos,length)"

   if { "$out(sub,$pos,ch)" == "01" } {
      IKE_print_sa_payload $out(sub,$pos,data)
   } elseif { "$out(sub,$pos,ch)" == "0d" } {
      IKE_print_vendor_id $out(sub,$pos,data)
   }
}



proc ISAKMP_payload_type { type } {

   set p [ expr 0x$type ]

   if { "$p" == "0" } {
     set str "NONE"
   } elseif { "$p" == "1" } {
     set str "Security Association"
   } elseif { "$p" == "2" } {
     set str "Proposal"
   } elseif { "$p" == "3" } {
     set str "Transform"
   } elseif { "$p" == "4" } {
     set str "Key Exchange"
   } elseif { "$p" == "5" } {
     set str "Identification"
   } elseif { "$p" == "6" } {
     set str "Certificate"
   } elseif { "$p" == "7" } {
     set str "Certificate Request"
   } elseif { "$p" == "8" } {
     set str "Hash"
   } elseif { "$p" == "9" } {
     set str "Signature"
   } elseif { "$p" == "10" } {
     set str "Nonce"
   } elseif { "$p" == "11" } {
     set str "Notification"
   } elseif { "$p" == "12" } {
     set str "Delete"
   } elseif { "$p" == "13" } {
     set str "Vendor ID"
   } elseif { "$p" >= "128" } {
     set str "Private ($p)"
   } else {
     set str "Unknown type '$p'"
   }


   return $str

#                    NONE                           0
#                    Security Association (SA)      1
#                    Proposal (P)                   2
#                    Transform (T)                  3
#                    Key Exchange (KE)              4
#                    Identification (ID)            5
#                    Certificate (CERT)             6
#                    Certificate Request (CR)       7
#                    Hash (HASH)                    8
#                    Signature (SIG)                9
#                    Nonce (NONCE)                 10
#                    Notification (N)              11
#                    Delete (D)                    12
#                    Vendor ID (VID)               13
#                    RESERVED                   14 - 127
#                    Private USE               128 - 255

   return $str
}







