#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: net.tcl,v 1.14 2002/11/19 23:16:46 joe Exp $
#

package provide pktsrc 1.3

proc iptohex { inaddr } {

    if [ catch {
       regexp {([0-9]*).([0-9]*).([0-9]*).([0-9]*)} $inaddr x a b c d
       set addr [list $a $b $c $d]

       set outaddr ""
       foreach x $addr {
	  set portion [format "%02x" $x]
	  set outaddr "$outaddr$portion"
       }
    } reason ] {

       error "failed to convert ip address '$inaddr' to hex"

    }

    return $outaddr
}

proc hextoip { inaddr } {

    if [ catch {

      set a [string range $inaddr 0 1]
      set b [string range $inaddr 2 3]
      set c [string range $inaddr 4 5]
      set d [string range $inaddr 6 7]
      set outaddr [ format "%u.%u.%u.%u" 0x$a 0x$b 0x$c 0x$d ]

    } reason ] {

      error "failed to convert hex string '$inaddr' to ip address"

    }

    return $outaddr
}

proc hextomac { inaddr } {

    set outaddr "failed to convert MAC '$inaddr'"

    catch {
      set a [string range $inaddr 0 1]
      set b [string range $inaddr 2 3]
      set c [string range $inaddr 4 5]
      set d [string range $inaddr 6 7]
      set e [string range $inaddr 8 9]
      set f [string range $inaddr 10 11]
      set outaddr [ format "%s:%s:%s:%s:%s:%s" $a $b $c $d $e $f ]
    }

    return $outaddr
}

proc mactohex { mac } {

    if [ catch {
       regexp {([0-9,a-f,A-F]):([0-9,a-f,A-F]*):([0-9,a-f,A-F]*):([0-9,a-f,A-F]*):([0-9,a-f,A-F]*):([0-9,a-f,A-F]*)} \
         $mac x a b c d e f

       set addr [list $a $b $c $d $e $f]

       set outaddr ""
       foreach x $addr {
	  set portion [format "%02x" 0x$x]
	  set outaddr "$outaddr$portion"
       }
    } reason ] {

       error "failed to convert mac address '$mac' to hex\n$reason"

    }

    return $outaddr
}

proc nextip { ip } {

    set hex [ iptohex $ip ]
    set nextip [ format "%8.8x" [ expr 0x$hex + 1 ]]
    return [ hextoip $nextip ]
}

proc maskbits { mask } {

    set maskhex [ iptohex $mask ]
    set count 0

    while { "$maskhex" != "00000000" } {
       incr count
       set maskhex [ format "%8.8x" [ expr 0x$maskhex << 1 ]]
    }
 
    return $count
}

#
# -- incr a 32 bit hex number in ascii hex format

proc incrhex { hex } {

    set result [ format "%8.8x" [ expr 0x$hex + 1 ]]
    return $result

}

#
# -- incr a 16 bit hex number in ascii hex format

proc incrhex16 { hex } {

    set result [ format "%4.4x" [ expr 0x$hex + 1 ]]
    return $result

}

proc hextoascii { string } {
    set len [ string length $string ]
    set ascii ""

    if { [ expr $len % 2 ] == 1 } {
        error "even size string required for hextoascii"
    } else {
	for { set i 0 } { $i < $len } { incr i 2 } {
            set data [ string range $string $i [ expr $i + 1 ] ]
            set ascii "$ascii[format "%c" 0x$data]"
        }
    }
    return $ascii
}

proc padhex { hexstring size } {

   set returnStr $hexstring

   set sizeT [ expr $size * 2 ]
   set length [ string length $hexstring ]

   if { $length < $sizeT } {
      set pad [ expr $sizeT - $length ]
      set padStr [ string repeat 0 $pad ]
      set returnStr "$hexstring$padStr"
   }

   return $returnStr
}

#
# classful_network
#
# -- return classful networks

proc classful_network { network } {

   set iphex [ iptohex $network ]
   set orig $iphex

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {

      # class A

      set net [ format "%8.8x" [ expr 0x$orig & 0xFF000000 ]] 
      return [ hextoip $net ]
   }
 
   set iphex [ format "%8.8x" [ expr 0x$iphex << 1 ]]

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {

      # class B

      set net [ format "%8.8x" [ expr 0x$orig & 0xFFFF0000 ]] 
      return [ hextoip $net ]
   }
 
   set iphex [ format "%8.8x" [ expr 0x$iphex << 1 ]]

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {

      # class C

      set net [ format "%8.8x" [ expr 0x$orig & 0xFFFFFF00 ]] 
      return [ hextoip $net ]
   }
 
   error "pktsrc: classful_network '$ip' not class A, B, or C"

}

proc classful_mask { mask } {
   set iphex [ iptohex $mask ]

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {
      return 255.0.0.0
   }
 
   set iphex [ format "%8.8x" [ expr 0x$iphex << 1 ]]

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {
      return 255.255.0.0
   }
 
   set iphex [ format "%8.8x" [ expr 0x$iphex << 1 ]]

   if { [ expr 0x$iphex & 0x80000000 ] == 0 } {
      return 255.255.255.0
   }
 
   error "pktsrc: classful_network '$ip' not class A, B, or C"

}

#
# -- return network address

proc ip_network_address { ip mask } {

   set host [ iptohex $ip ]
   set mask [ iptohex $mask ]

   # -- get network portion
   set net [ format "%8.8x" [ expr 0x$host & 0x$mask ]]

   return [ hextoip $net ]

}

#
# -- return broadcast address

proc ip_broadcast_address { ip mask } {

   set host [ iptohex $ip ]
   set mask [ iptohex $mask ]

   # -- get broadcast
   set broadcast [ format "%8.8x" [ expr (0x$host & 0x$mask) | (~0x$mask) ]]

   return [ hextoip $broadcast ]
}

#
# ip_range_contains
#
# return true if address fails within the start and stop ip address ranges
#

proc ip_range_contains { start stop address } {

   set starthex [ iptohex $start ]
   set stophex [ iptohex $stop ]
   set ip [ iptohex $address ]

   return [ expr (0x$stophex >= 0x$ip) && (0x$ip >= 0x$starthex) ] 

}

#
# wireless_interface
#
# helper proc for determining wireless interfaces
#

proc wireless_interface { interface } {
 
    set dev [ string range $interface 0 3 ]
 
    if { "$dev" == "wlan" } {
       return 1
    }
 
    return 0
}

# -- event related tools

#
# Event_wait does a non blocking wait for 'length' msecs

proc Event_wait { length { message "" }} {
    global ___event_wait_flag

    set ___event_wait_flag 0

    # -- set up an async timer callback to touch the var
    after $length "
       global ___event_wait_flag
       set ___event_wait_flag 1
    "

    if { "$message" != "" } {
       pktsrcLog $message
    }

    # -- wait for the var to change
    vwait ___event_wait_flag
}

#
# Event_wait_packet does a non blocking wait for 'length' msecs
# It returns earlier if the Event_got_packet is called.
#
# Only one Event_wait_packet call can be running at a time. None of the
# 'callback' type procs should ever call Event_wait_packet. This
# call is intended for 'blocking' api calls.
#

set ___event_packet_wait_flag -1

proc Event_wait_packet { length { message "" }} {
    global ___event_packet_wait_flag
    global ___event_packet_wait_flag_id

    
    if { "$___event_packet_wait_flag" != "-1" } {
       error "pktsrc fatal error: Event_wait_packet called, but already running"
       set ___event_packet_wait_flag 0
    }

    # -- set up an async timer callback to touch the var
    set ___event_packet_wait_flag_id [ after $length "
       global ___event_packet_wait_flag
       set ___event_packet_wait_flag 1
    "]

    if { "$message" != "" } {
       pktsrcLog $message
    }

    # -- wait for the var to change

    vwait ___event_packet_wait_flag

    # -- return the trigger code
 
    #  1 -> timeout
    #  2 -> packet received

    set retCode $___event_packet_wait_flag    
    
    set ___event_packet_wait_flag -1    
    return $retCode 

}

proc Event_got_packet { } {
    global ___event_packet_wait_flag
    global ___event_packet_wait_flag_id

    set ___event_packet_wait_flag 2
    after cancel $___event_packet_wait_flag_id

}

proc pktsrc_copy_packet { packet_in packet_out } {
    upvar $packet_in inp
    upvar $packet_out outp

    # -- make sure packet_in is a packet/array
    if {![array exists inp]} {
	error "pktsrc_copy_packet: \"$packet_in\" is not a packet"
    }

    # -- copy all the array members of inp
    foreach name [array names inp *] {
       set outp($name) $inp($name)
    }
}




