/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (c) 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  http://qacafe.com
 *
 *  $Id: pktsrc.c,v 1.10 2002/11/22 15:59:51 joe Exp $
 *
 */

#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <linux/if_ether.h>

#include <pcap.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>

#include <netinet/in.h>

#include <errno.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define NEED_PCAP_SOURCE
#ifdef NEED_PCAP_SOURCE
#include "pcap-int.h"
/* #include "gnuc.h" */
#endif

#ifdef HAVE_OS_PROTO_H
#include "os-proto.h"
#endif

#ifndef MSG_TRUNC
#warning MSG_TRUNC is not defined: your libc is broken. Please, fix it.
#define MSG_TRUNC 0
#endif

#ifdef PACKET_TRECV
#include <linux/poll.h>
#include <linux/mman.h>
#endif


char ebuf[PCAP_ERRBUF_SIZE];


void hex_dump ( unsigned char *buffer, int length );

int 
pktsrc_open( pcap_t **mypd, char *device, char *ebuf )
{
  int i, snaplen = 1518;
  struct bpf_program fcode;
  bpf_u_int32 localnet = 0, netmask = 0;
  int Oflag=1, datalink = 0;
  pcap_t *pd = 0;

  pd = pcap_open_live(device, snaplen, -1, 1000, ebuf);

  *mypd = pd;

  if (pd == 0) {
     return -1;
  }

  i = pcap_snapshot(pd);
 
  if (snaplen < i) {
      sprintf(ebuf,"snaplen raised from %d to %d\n", snaplen, i);
      snaplen = i;
  }

  /*
   * Build a filter
   */

   if (pcap_compile(pd, &fcode, "", Oflag, netmask) < 0) {
       sprintf(ebuf,"%s\n", pcap_geterr(pd));
       return -1;
   }

  /*
   * Set the filter
   */

   if (pcap_setfilter(pd, &fcode) < 0) {
	sprintf(ebuf,"%s\n", pcap_geterr(pd));
	return -1;
   }

   return 0;

}

int
pktsrc_write(pcap_t *pd, unsigned char *buf,int count)
{
   int i = 0;
   struct sockaddr_ll sa;
   struct ifreq interface;

   sa.sll_family = AF_PACKET;

   strcpy(interface.ifr_name, pd->md.device);

   i = ioctl( pcap_fileno(pd), SIOCGIFINDEX, &interface);
 
   if ( i < 0 ) {
      perror("Could not find interface index");
      return (-1);
   }

   sa.sll_ifindex = interface.ifr_ifindex;
   sa.sll_protocol = ETH_P_ALL;

   i = sendto( pcap_fileno(pd), buf, count, 0, (struct sockaddr *)&sa, sizeof(sa));

   if ( i < 0 )
   {
      return (-1);
   }

   if (i != count) {
      perror("write");
      return -1;
   } else {
      return i;
   }
}             



int pktsrc_send_packet( int dump, pcap_t *pd, unsigned char *packet, int size )
{
    int s = 0;

    if (dump == 1 ) {
       printf("\n>>Sending packet on %s, length %d:\n", pd->md.device, size);
       hex_dump( packet, size );
       printf("\n");
       fflush(stdout);
    }

    s = pktsrc_write(pd, packet,size);
    return s;
}

int
pktsrc_read(pcap_t *handle, int cnt, char **buffer, int *length, struct timeval *t)
{
	u_char			*bp;
	int			offset;
        int                     wireless = 0;
#ifdef HAVE_PF_PACKET_SOCKETS
	struct sockaddr_ll	from;
	struct sll_header	*hdrp;
#else
	struct sockaddr		from;
#endif
	socklen_t		fromlen;
	int			packet_len, caplen;
	struct pcap_pkthdr	pcap_header;

#ifdef HAVE_PF_PACKET_SOCKETS
	/*
	 * If this is a cooked device, leave extra room for a
	 * fake packet header.
	 */
	if (handle->md.cooked)
		offset = SLL_HDR_LEN;
	else
		offset = 0;
#else
	/*
	 * This system doesn't have PF_PACKET sockets, so it doesn't
	 * support cooked devices.
	 */
	offset = 0;
#endif

	/* Receive a single packet from the kernel */

	bp = handle->buffer + handle->offset;
	do {
		fromlen = sizeof(from);
		packet_len = recvfrom(
			handle->fd, bp + offset,
			handle->bufsize - offset, MSG_TRUNC,
			(struct sockaddr *) &from, &fromlen);
	} while (packet_len == -1 && errno == EINTR);

	/* Check if an error occured */

	if (packet_len == -1) {
		if (errno == EAGAIN)
			return 0;	/* no packet there */
		else {
			snprintf(handle->errbuf, sizeof(handle->errbuf),
				 "recvfrom: %s", pcap_strerror(errno));
			return -1;
		}
	}

#ifdef HAVE_PF_PACKET_SOCKETS
	/*
	 * If this is from the loopback device, reject outgoing packets;
	 * we'll see the packet as an incoming packet as well, and
	 * we don't want to see it twice.
	 *
	 * We can only do this if we're using PF_PACKET; the address
	 * returned for SOCK_PACKET is a "sockaddr_pkt" which lacks
	 * the relevant packet type information.
	 */
	if (!handle->md.sock_packet &&
	    from.sll_ifindex == handle->md.lo_ifindex &&
	    from.sll_pkttype == PACKET_OUTGOING)
		return 0;
#endif

#ifdef HAVE_PF_PACKET_SOCKETS
	/*
	 * If this is a cooked device, fill in the fake packet header.
	 */
	if (handle->md.cooked) {
		/*
		 * Add the length of the fake header to the length
		 * of packet data we read.
		 */
		packet_len += SLL_HDR_LEN;

		hdrp = (struct sll_header *)bp;

		/*
		 * Map the PACKET_ value to a LINUX_SLL_ value; we
		 * want the same numerical value to be used in
		 * the link-layer header even if the numerical values
		 * for the PACKET_ #defines change, so that programs
		 * that look at the packet type field will always be
		 * able to handle DLT_LINUX_SLL captures.
		 */
		switch (from.sll_pkttype) {

		case PACKET_HOST:
			hdrp->sll_pkttype = htons(LINUX_SLL_HOST);
			break;

		case PACKET_BROADCAST:
			hdrp->sll_pkttype = htons(LINUX_SLL_BROADCAST);
			break;

		case PACKET_MULTICAST:
			hdrp->sll_pkttype = htons(LINUX_SLL_MULTICAST);
			break;

		case PACKET_OTHERHOST:
			hdrp->sll_pkttype = htons(LINUX_SLL_OTHERHOST);
			break;

		case PACKET_OUTGOING:
			hdrp->sll_pkttype = htons(LINUX_SLL_OUTGOING);
			break;

		default:
			hdrp->sll_pkttype = -1;
			break;
		}

		hdrp->sll_hatype = htons(from.sll_hatype);
		hdrp->sll_halen = htons(from.sll_halen);
		memcpy(hdrp->sll_addr, from.sll_addr,
		    (from.sll_halen > SLL_ADDRLEN) ?
		      SLL_ADDRLEN :
		      from.sll_halen);
		hdrp->sll_protocol = from.sll_protocol;
	}
#endif

	/*
	 * XXX: According to the kernel source we should get the real
	 * packet len if calling recvfrom with MSG_TRUNC set. It does
	 * not seem to work here :(, but it is supported by this code
	 * anyway.
	 * To be honest the code RELIES on that feature so this is really
	 * broken with 2.2.x kernels.
	 * I spend a day to figure out what's going on and I found out
	 * that the following is happening:
	 *
	 * The packet comes from a random interface and the packet_rcv
	 * hook is called with a clone of the packet. That code inserts
	 * the packet into the receive queue of the packet socket.
	 * If a filter is attached to that socket that filter is run
	 * first - and there lies the problem. The default filter always
	 * cuts the packet at the snaplen:
	 *
	 * # tcpdump -d
	 * (000) ret      #68
	 *
	 * So the packet filter cuts down the packet. The recvfrom call
	 * says "hey, it's only 68 bytes, it fits into the buffer" with
	 * the result that we don't get the real packet length. This
	 * is valid at least until kernel 2.2.17pre6.
	 *
	 * We currently handle this by making a copy of the filter
	 * program, fixing all "ret" instructions with non-zero
	 * operands to have an operand of 65535 so that the filter
	 * doesn't truncate the packet, and supplying that modified
	 * filter to the kernel.
	 */

	caplen = packet_len;
	if (caplen > handle->snapshot)
		caplen = handle->snapshot;

	/* Run the packet filter if not using kernel filter */
	if (!handle->md.use_bpf && handle->fcode.bf_insns) {
		if (bpf_filter(handle->fcode.bf_insns, bp,
		                packet_len, caplen) == 0)
		{
			/* rejected by filter */
			return 0;
		}
	}

	/* Fill in our own header data */

	if (ioctl(handle->fd, SIOCGSTAMP, &pcap_header.ts) == -1) {
		snprintf(handle->errbuf, sizeof(handle->errbuf),
			 "ioctl: %s", pcap_strerror(errno));
		return -1;
	}
	pcap_header.caplen	= caplen;
	pcap_header.len		= packet_len;


        *length = pcap_header.caplen;
        *buffer = bp; 
        *t = pcap_header.ts;

	/*
	 * Count the packet.
	 *
	 * Arguably, we should count them before we check the filter,
	 * as on many other platforms "ps_recv" counts packets
	 * handed to the filter rather than packets that passed
	 * the filter, but if filtering is done in the kernel, we
	 * can't get a count of packets that passed the filter,
	 * and that would mean the meaning of "ps_recv" wouldn't
	 * be the same on all Linux systems.
	 *
	 * XXX - it's not the same on all systems in any case;
	 * ideally, we should have a "get the statistics" call
	 * that supplies more counts and indicates which of them
	 * it supplies, so that we supply a count of packets
	 * handed to the filter only on platforms where that
	 * information is available.
	 *
	 * We count them here even if we can get the packet count
	 * from the kernel, as we can only determine at run time
	 * whether we'll be able to get it from the kernel (if
	 * HAVE_TPACKET_STATS isn't defined, we can't get it from
	 * the kernel, but if it is defined, the library might
	 * have been built with a 2.4 or later kernel, but we
	 * might be running on a 2.2[.x] kernel without Alexey
	 * Kuznetzov's turbopacket patches, and thus the kernel
	 * might not be able to supply those statistics).  We
	 * could, I guess, try, when opening the socket, to get
	 * the statistics, and if we can not increment the count
	 * here, but it's not clear that always incrementing
	 * the count is more expensive than always testing a flag
	 * in memory.
	 */
	handle->md.stat.ps_recv++;

	return 1;
}


int
get_stats(pcap_t *p, struct pcap_stat *data )
{
    return pcap_stats( p, data );
}

void 
hex_dump ( unsigned char *buffer, int length )
{
    int i = 0;
    for ( i = 0; i < length; i++ ) {
        if ( i % 16 == 0 ){
            printf("\n");
        } else {
            printf(":");
        }
        printf("%.2x", buffer[i] );
    }
    printf("\n");
}

int 
dump_packet( struct pcap_pkthdr *h1, u_char *bp )
{
    int i = 0;
    for ( i = 0; i < h1->caplen; i++ ) {
        if ( i%16 == 0 ){
            printf("\n");
        } else {
            printf(":");
        }
        printf("%.2x", bp[i] );
    }
    printf("\n");
}

void 
set_dest_mac( char *packet, char *mac )
{
    memcpy( packet, mac, 6);
}

int 
ipcksum(char *data, int length )
{
    unsigned int sum = 0;
    unsigned short *s = (unsigned short *)data;

    while ( length > 0 ) {
       sum += *s; 
       s++;
       length -= 2;
    }

    while ( sum > 0xffff ) {
        sum = ( sum & 0xffff ) + (sum >> 16 );
    }

    sum = ~sum & 0xffff;
 
    return sum; 
} 

unsigned short
tcp_udp_cksum (char *data, int len)
{
    unsigned short *sptr;
    unsigned i;
    unsigned int tcksum = 0;
    unsigned char proto;

    /* collect the addresses */
    sptr = (unsigned short *)(data + 26); /* start of addresses in packet */
    for (i=0; i<4; ++i) {
        tcksum += *sptr++;
    }

    proto = *((unsigned char *)(data + 23));
    tcksum += htons((int)proto + len);

    sptr = (unsigned short *)(data + 34); /* start of tcp data in packet */
    data = (data + 34);
  
    if (len % 2) {
        data[len] = 0;
        len += 1;
    }
    len >>=1;

    for (i=0; i<len; ++i) {
        tcksum += *sptr++;
    }
  
    tcksum = (tcksum >> 16) + (tcksum & 0xffff);
    tcksum += (tcksum >> 16);

    return (short)(~tcksum & 0xffff);
}


/* This is the fletcher checksum from rfc 1008
 */

#define MODX 4102 
int fletcherCksum( unsigned char *mess, int len, int k )
{ 

  /*
    unsigned char mess[]     the TPDU to be checksummed 
    int len,
            k;               position of first checksum octet
                             as an offset from mess[0]
   */


      int ip, iq, ir, c0, c1;
      unsigned char *p,*p1,*p2,*p3 ;
      unsigned short cksum;

      p = mess ; p3 = mess + len ; 

      /* insert zeros for checksum octets */
      /* if ( k > 0) { mess[k-1] = 0x00 ; mess[k] = 0x00 ; } */

      c0 = 0 ; c1 = 0 ; p1 = mess ; 

      while (p1 < p3)    /* outer sum accumulation loop */
      {
               p2 = p1 + MODX ; if (p2 > p3) p2 = p3 ;

               /*  inner sum accumulation loop */
               for (p = p1 ; p < p2 ; p++) { 
                  c0 = c0 + (*p) ; 
                  c1 = c1 + c0 ;
               }

               /* adjust accumulated sums to mod 255 */
               c0 = c0%255 ; 
               c1 = c1%255 ; 
               p1 = p2 ;

       }

       ip = (c1 << 8) + c0 ;     /* concatenate c1 and c0 */

       if (k > 0)  {     /* compute and insert checksum octets */

          iq = ((len-k)*c0 - c1)%255 ; 
          if (iq <= 0) iq = iq + 255 ; /* mess[k-1] = iq ; */
          ir = (510 - c0 - iq) ; 
          if (ir > 255) ir = ir - 255 ; /* mess[k] = ir ; */

       } 

       /* build the cksum rather than insert in the buffer */
       cksum = (unsigned short)((iq << 8 ) | ir );       
      
       return(cksum) ; 
} 


/* that's it */





