#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ppp-chap.tcl,v 1.5 2002/11/26 17:21:17 joe Exp $
#
#
# pktsrc PPP/CHAP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc PPP_CHAP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {
       if [ catch { PPP_CHAP_decode $packet chap } reason ] {
          pktsrcWarn s "Unable to fully parse PPP CHAP packet"
       } else {
          PPP_CHAP_print chap
       }
       # -- nothing above chap, return
       return
   }

   # -- remember this packet in the 'chap' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s CHAP chap ]
   
   # -- now decode the packet into the 'chap' variable
   if [ catch { PPP_CHAP_decode $packet chap } reason ] {
      pktsrcWarn s "Unable to fully parse PPP CHAP packet"
      return
   }

   # -- check for callbacks 

   if [ info exists s(chapType,$chap(code)) ] {

      # -- drive any input callbacks
      foreach c $s(chapType,$chap(code),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   } else {
      # no callback
   }
}

#
#
# PPP_CHAP_init - set up any deaults for CHAP
#

proc PPP_CHAP_init { stack } {
   upvar $stack s

   # -- initial CHAP options

   set s(CHAP,user) qacafe
   set s(CHAP,password) hello
   set s(CHAP,mode) off
   set s(CHAP,type) chap
   set s(CHAP,secretLen) 16

   # -- if the identifier exists already, just bump it
   if [ info exists s(CHAP,identifier) ] {
      incr s(CHAP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(CHAP,identifier) > 255 } {
          set s(CHAP,identifier) 1
       }         

   } else {

      # -- otherwise start at one
      set s(CHAP,identifier) 1
   }

   set s(CHAP,state) down
}

proc PPP_CHAP_decode { packet CHAP } {
   upvar $CHAP out

   # -- parse out all the fields we find interesting ...
   set out(code)         [ string range $packet 0 1 ]
   set out(id)           [ string range $packet 2 3 ]
   set out(length)       [ string range $packet 4 7 ]

   # -- check for tags to decode   
   set len [expr [ string length $packet ] / 2 ]

   set out(options) 0
   set code [ expr 0x$out(code) ] 

   if { "$code" == "1" || "$code" == "2" } {
      set out(value-len)       [ string range $packet 8 9 ]
      set len [ expr  0x$out(value-len) ]
      set myend [ expr 10 + ($len * 2 ) - 1]
      set out(value) [ string range $packet 10 $myend ]

      incr myend
      # +++joem check length 
      set out(name)       [ string range $packet $myend end ]
   } elseif { "$code" == "3" || "$code" == "4" } {

      # -- really check length
      set out(message) [ string range $packet 8 [expr $len * 2 - 1 ]]
   }
}

proc PPP_CHAP_send { stack packet protocol } {

   upvar $stack s

   error "PPP_CHAP_send: not implemented"
}



proc PPP_CHAP_print { CHAP } {
   upvar $CHAP out

   set code [ expr 0x$out(code) ]

   puts "CHAP: ---- CHAP Packet ----"
   puts "CHAP: "
   puts "CHAP: Code                      = [PPP_CHAP_get_code $out(code)] (0x$out(code))"
   puts "CHAP: Identifier                = [ expr 0x$out(id)]"
   puts "CHAP: Length                    = [ expr 0x$out(length)]"

   set code [ expr 0x$out(code) ] 

   if { "$code" == "1" || "$code" == "2" } {
      set name [ pktsrc ascii2text $out(name) ]
      puts "CHAP: Value Length              = [expr $out(value-len)]"
      puts "CHAP: Value                     = $out(value)"
      puts "CHAP: Name                      = $name"
   } elseif { "$code" == "03" || "$code" == "04" } {
      if { "$out(message)" != "" } {
         puts "CHAP: Message                   = 0x$out(message)"
      } else {
         puts "CHAP: Message                   = \"\""
      }

   }

   puts "CHAP:"
   puts "CHAP: ---- End CHAP Packet ----"

}

proc PPP_CHAP_get_code { code } {
  set str "Unknown"

  set code [ expr 0x$code ]

  if { "$code" == "1" } {
     set str  "Challenge"
  } elseif { "$code" == "2" } {
      set str "Response"
  } elseif { "$code" == "3" } {
      set str "Success"
  } elseif { "$code" == "4" } {
      set str "Failure"
  }

  return $str

}

proc PPP_CHAP_get_auth { type } {
  set str "Unknown"

  if { "$type" == "c22305" } {
     set str  "CHAP"
  } elseif { "$type" == "c22380" } {
     set str  "MSCHAP"
  } elseif { "$type" == "c023" } {
      set str "PAP"
  }

  return $str

}

proc CHAP_set_user { stack user password } {
   upvar $stack s

   set s(CHAP,user) $user
   set s(CHAP,password) $password

   pktsrcInfo s "Setting PPP CHAP user to $user with password $password"

}

proc CHAP_get_user { stack } {
   upvar $stack s
   return $s(CHAP,user)
}

proc CHAP_get_password { stack } {
   upvar $stack s
   return $s(CHAP,password)
}


proc CHAP_client_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the CHAP packet from the stack

   set chapState $s(current,CHAP)
   upvar #$s(current,CHAP,level) $chapState chap

   pktsrcInfo s "Received CHAP [ PPP_CHAP_get_code $chap(code) ]"

   if { $chap(code) == "01" } {

      pktsrcInfo s "Ignoring unexpected CHAP challenge"

   } elseif { $chap(code) == "02" } {

      pktsrcInfo s "CHAP response value is '$chap(value)'"
      
      set passwd [ pktsrc text2ascii $s(CHAP,password) ]

      # -- build my MD5 over id, secret, response value
      set secret $s(CHAP,currentSecret)

      set mymd5 [ pktsrc md5 ${chap(id)}${passwd}${secret} ]

      pktsrcInfo s "My computed response value is '$mymd5'"

      if { "$mymd5" == "$chap(value)" } {

         # -- validate
         pktsrcInfo s "Accepting CHAP authentication from PPP client"
         set s(CHAP,state) up
         CHAP_send_success s chap

      } else {

         pktsrcWarn s "CHAP authentication from PPP client failed"
         CHAP_send_failure s chap
         set s(CHAP,state) down
      }

   } elseif { $chap(code) == "03" } {

      pktsrcInfo s "CHAP authentication was accepted"

   } elseif { $chap(code) == "04" } {

      # -- configuration nak'd
      pktsrcWarn s "Ignoring unexpected CHAP Failure message"

   }
}

proc CHAP_send_challenge { stack { options "" }} {
   upvar $stack s

   if { "$options" != "" } {
      error "CHAP_send_config_request: options not supported yet"
   }

   pktsrcInfo s "Sending CHAP [ PPP_CHAP_get_code 01 ]"
   
   set usertext [ pktsrc text2ascii $s(CHAP,user) ]

   # -- get a secret
   set secret [ PPP_CHAP_get_secret s ]

   set s(CHAP,currentSecret) $secret

   # -- send challenge
   set secretLen [ expr ([string length $secret]/2)]

   set len [ expr 5 + (([string length $usertext ])/2) + $secretLen ]
   set lenhex [ format "%4.4x" $len ]
   set slhex [ format "%2.2x" $secretLen ]
   set packet "0101${lenhex}${slhex}${secret}${usertext}"

   # -- send as CHAP
   PPP_send s $packet c223

   # -- bump the identifier for the next challenge
   PPP_CHAP_next_identifier s
}

proc CHAP_send_success { stack chap } {
   upvar $stack s
   upvar $chap l

   pktsrcInfo s "Sending CHAP [ PPP_CHAP_get_code 03 ]"

   # -- build config ack for this id
   set packet "03${l(id)}0004"

   # -- send as a CHAP packet to PPP
   PPP_send s $packet c223
}

proc CHAP_send_failure { stack chap } {
   upvar $stack s
   upvar $chap l

   pktsrcInfo s "Sending CHAP [ PPP_CHAP_get_code 04 ]"

   # -- build config nack for this id with this set of options
   set packet "04${l(id)}0004"

   # -- send as a CHAP packet to PPP
   PPP_send s $packet c223
}

proc PPP_CHAP_start { stack { mode client } } {
   upvar $stack s
   # -- build a configure request

   pktsrcInfo s "Starting CHAP"

   # -- Add callbacks for all CHAP packets
   Stack_add_callback s chapType 01 CHAP_client_input
   Stack_add_callback s chapType 02 CHAP_client_input
   Stack_add_callback s chapType 03 CHAP_client_input
   Stack_add_callback s chapType 04 CHAP_client_input

   # -- send config request
   if { $mode == "server" } {
      CHAP_send_challenge s
   }
}

proc PPP_CHAP_set_secret_len { stack length } {
   upvar $stack s

   if { $length > 256 } {
       error "PPP_CHAP_set_secret_len: max length is 255"
   }

   set s(CHAP,secretLen) $length
}

proc PPP_CHAP_get_secret { stack } {
   upvar $stack s

   # -- this needs to be random
   set charhex [string range [format "%8.8x" [ clock clicks ]] 6 7 ]
   set secret [ string repeat $charhex $s(CHAP,secretLen) ]

   return $secret
}

proc PPP_CHAP_next_identifier { stack } {
   upvar $stack s

   if [ info exists s(CHAP,identifier) ] {
      incr s(CHAP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(CHAP,identifier) > 255 } {
          set s(CHAP,identifier) 1
       }         
   }
}

proc CHAP_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
     Event_got_packet
   }
}

proc CHAP_get_packet { stack type packet timeOut } {
   upvar $stack s
   set ret -1

   # -- add a callback for packets
   Stack_add_callback s chapType $type CHAP_generic_input

   pktsrcInfo s "waiting up to $timeOut msecs for [ PPP_CHAP_get_code $type ] packet"
   set result [ Event_wait_packet $timeOut ]

   if { "$result" == "2" } {

      # -- get the contents of the CHAP packet from the stack
      set chapState $s(current,CHAP)
      upvar #$s(current,CHAP,level) $chapState chap
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet chap p
      set ret 0

   } elseif { "$result" == "1" } {
      pktsrcInfo s "timeout waiting for [ PPP_CHAP_get_code $type ] packet" 
      set ret -1
   } else {
      set ret -1
   }

   Stack_remove_callback s chapType $type CHAP_generic_input

   return $ret
}

proc CHAP_get_state { stack } {
   upvar $stack s

   return $s(CHAP,state)

}








