#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ppp-lcp.tcl,v 1.7 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc PPP/LCP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc PPP_LCP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {
      if [ catch { PPP_LCP_decode $packet lcp } reason ] {
         pktsrcWarn s "Unable to fully parse PPP LCP packet"
      } else {
         PPP_LCP_print lcp
      }

      # -- nothing above lcp, return
      return

   }

   # -- remember this packet in the 'lcp' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s LCP lcp ]
   
   # -- now decode the packet into the 'lcp' variable
   if [ catch { PPP_LCP_decode $packet lcp } reason ] {
      pktsrcWarn s "Unable to fully parse PPP LCP packet"
      return
   }

   # -- check for callbacks 

   if [ info exists s(lcpType,$lcp(code)) ] {

      # -- drive any input callbacks
      foreach c $s(lcpType,$lcp(code),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   } else {
      # no callback
   }
}

#
# PPP_LCP_reset
#

proc PPP_LCP_reset { stack } {
   upvar $stack s


   set s(LCP,magic-number) [ format "%8.8x" [ clock clicks ]]

   # -- if the identifier exists already, just bump it
   if [ info exists s(LCP,identifier) ] {
      incr s(LCP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(LCP,identifier) > 255 } {
          set s(LCP,identifier) 1
       }         

   } else {

      # -- otherwise start at one
      set s(LCP,identifier) 1
   }

   # -- make sure we are down
   set s(LCP,state) down
   set s(LCP,renegotiate) 0

}

proc PPP_LCP_next_identifier { stack } {
   upvar $stack s

   if [ info exists s(LCP,identifier) ] {
      incr s(LCP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(LCP,identifier) > 255 } {
          set s(LCP,identifier) 1
       }         
   }
}

#
#
# PPP_LCP_init - set up any deaults for LCP
#

proc PPP_LCP_init { stack } {
   upvar $stack s

   # -- initial LCP options

   set s(LCP,auth) PAP
   set s(LCP,mode) client
   set s(LCP,magic-number) [ format "%8.8x" [ clock clicks ]]

   set s(LCP,mru) [ format "%4.4x" 1500 ]
   # -- create a set option for this

   # -- if the identifier exists already, just bump it
   if [ info exists s(LCP,identifier) ] {
      incr s(LCP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(LCP,identifier) > 255 } {
          set s(LCP,identifier) 1
       }         

   } else {

      # -- otherwise start at one
      set s(LCP,identifier) 1
   }

   set s(LCP,state) down
   set s(LCP,echoMode) on
   set s(LCP,renegotiate) 0
}

# 00:0b:0b:00:00:01:00:20:78:d2:bf:a6:88:64:11:00
# 00:01:00:14:c0:21:02:02:00:12:01:04:05:dc:05:06
# 2b:a4:38:c5:00:00:00:00:00:00:00:00:00:00:00:00
# 00:00:00:00:00:00:00:00:00:00:00:00

proc PPP_LCP_decode { packet LCP } {
   upvar $LCP out

   # -- parse out all the fields we find interesting ...
   set out(code)         [ string range $packet 0 1 ]
   set out(id)           [ string range $packet 2 3 ]
   set out(length)       [ string range $packet 4 7 ]

   # -- check for tags to decode   
   set len [expr [ string length $packet ] / 2 ]

   set out(options) 0


   set code [ expr 0x$out(code) ] 

   # -- decode options for those codes that support them
   if { $code <= 4 } {

      if { [expr 0x$out(length)] > $len } {
         puts "ERROR: Missing LCP data"
   
      } elseif { $len > 8 } {
         PPP_LCP_decode_options $packet out
      }
   } else {

      # -- handle other codes here
      # +++joem, print proc needs to handle these separately as well

      if { $code == 9 || $code == 10 } {

         set out(magic-number)       [ string range $packet 8 15 ]

         # -- check for trailing data
         set len [ expr 0x$out(length) ]
	 if { $len > 8 } {
            # extra data included
            set myend [ expr 16 + ($len - 8)*2 - 1  ]
            set out(data) [ string range $packet 16 $myend ]
         } else {
            set out(data) ""
         }
      }
   }
}

proc PPP_LCP_decode_options { packet lcp } {
   upvar $lcp out

   set p 8

   set out(options) 0

   set count 0
   set totalLen 4

   while { 1 } {
     
      if { $totalLen >= [ expr 0x$out(length) ] } {
         break
      } 
      set tag [ string range $packet $p [expr $p + 1]]
      # puts "tag is $tag"

      incr p 2
      set length [ string range $packet $p [expr $p + 1]]
      # puts "length is $length"

      incr p 2

      set out(option,$count,tag) $tag
      set out(option,$count,length) $length
      set out(option,$count,data) \
          [ string range $packet $p [expr $p + 2 * 0x$length - 5 ]]
      incr p [ expr 0x$length * 2 - 4 ]

      # puts "data is $out(option,$count,data)"
      incr count

      incr totalLen [ expr 0x$length ]

      if { ![ expr 0x$length > 0 ] } {
         break
      }
   } 

   # -- set the total number of tags found
   set out(options) $count  
}

proc PPP_LCP_send { stack packet } {

   upvar $stack s

   # -- currently things are calling PPP_send directory
   error "not supported yet"

}



proc PPP_LCP_print { LCP } {
   upvar $LCP out

   set code [ expr 0x$out(code) ]

   puts "LCP: ---- LCP Packet ----"
   puts "LCP: "
   puts "LCP: Code                       = [PPP_LCP_get_code $out(code)] (0x$out(code))"
   puts "LCP: Identifier                 = [ expr 0x$out(id)]"
   puts "LCP: Length                     = [ expr 0x$out(length)]"


   if { $code == 9 || $code == 10 } {
         puts "LCP: Magic Number               = 0x$out(magic-number)"
       if { "$out(data)" != "" } {
         puts "LCP: Data                       = 0x$out(data)"
       } else {
         puts "LCP: Data                       = \"\" (no data)"
       }

   } else {

   puts "LCP: "
   puts "LCP: ---- Options ----"
   puts "LCP:"

   if { $out(options) > 0 } {
      PPP_LCP_print_options out
   } else {
      puts "LCP: No options found"
   }

   }
   puts "LCP:"
   puts "LCP: ---- End LCP Packet ----"

}

proc PPP_LCP_print_options { lcp } {
   upvar $lcp out

#
#         0       RESERVED
#         1       Maximum-Receive-Unit
#         3       Authentication-Protocol
#         4       Quality-Protocol
#         5       Magic-Number
#         7       Protocol-Field-Compression
#         8       Address-and-Control-Field-Compression
#

   for { set i 0 } { $i < $out(options) } { incr i } {

      # -- get the option tag in decimal

      set tag $out(option,$i,tag)
      set length $out(option,$i,length)
      set data $out(option,$i,data)

      if { "$tag" == "01" } {

         puts "LCP: MRU                        = [ expr 0x$data]"

      } elseif { "$tag" == "03" } {

         set endp [ expr ( $length - 2 ) * 2 - 1]
         set protocol [ string range $data 0 $endp  ]

         puts "LCP: Auth Protocol              = [ PPP_LCP_get_auth $protocol ] (0x$protocol)"

      } elseif { "$tag" == "05" } {

         puts "LCP: Magic Number               = 0x$data"

      } elseif { "$tag" == "07" } {

         puts "LCP: Protocol-Field-Compression"

      } else {

         # -- default option display

         puts "LCP: Option Type                = (0x$tag)"
         puts "LCP: Option Length              = $length"
      }
   }
}

proc PPP_LCP_get_code { code } {
  set str "Unknown"

  set code [ expr 0x$code ]

#
# from rfc 1661
#
#                1       Configure-Request
#                2       Configure-Ack
#                3       Configure-Nak
#                4       Configure-Reject
#                5       Terminate-Request
#                6       Terminate-Ack
#                7       Code-Reject
#                8       Protocol-Reject
#                9       Echo-Request
#                10      Echo-Reply
#                11      Discard-Request

  if { "$code" == "1" } {
     set str  "Configure-Request"
  } elseif { "$code" == "2" } {
      set str "Configure-Ack"
  } elseif { "$code" == "3" } {
      set str "Configure-Nak"
  } elseif { "$code" == "4" } {
      set str "Configure-Reject"
  } elseif { "$code" == "5" } {
      set str "Terminate-Request"
  } elseif { "$code" == "6" } {
      set str "Terminate-Ack"
  } elseif { "$code" == "7" } {
      set str "Code-Reject"
  } elseif { "$code" == "8" } {
      set str "Protocol-Reject"
  } elseif { "$code" == "9" } {
      set str "Echo-Request"
  } elseif { "$code" == "10" } {
      set str "Echo-Reply"
  } elseif { "$code" == "11" } {
      set str "Discard-Request"
  }

  return $str

}

proc PPP_LCP_get_auth { type } {
  set str "Unknown"

  if { "$type" == "c22305" } {
     set str  "CHAP"
  } elseif { "$type" == "c22380" } {
     set str  "MSCHAP"
  } elseif { "$type" == "c023" } {
      set str "PAP"
  }

  return $str

}

proc LCP_client_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the LCP packet from the stack

   set lcpState $s(current,LCP)
   upvar #$s(current,LCP,level) $lcpState lcp

   pktsrcInfo s "Received LCP [ PPP_LCP_get_code $lcp(code) ] (id [expr 0x$lcp(id)])"

   if { $lcp(code) == "01" } {

      # -- got configuration request, just sent ack
      LCP_send_config_ack s lcp

      # -- start NCP & PAP
      PPP_NCP_start s $s(LCP,mode)

      # -- see if PAP is enabled
      if { $s(PAP,mode) == "on" } { 
         PPP_PAP_start s $s(LCP,mode)
      }


      # -- if I am in server mode, send my config request now
      if { $s(LCP,mode) == "server" &&  $s(LCP,renegotiate) != "1" } {
          LCP_send_config_request s
      }

      set s(LCP,renegotiate) 0

   } elseif { $lcp(code) == "02" } {
      pktsrcInfo s "LCP configuration was accepted"
      set s(LCP,renegotiate) 0

      # -- we now consider this link up
      set s(LCP,state) up

      # -- start CHAP if enabled
      if { $s(CHAP,mode) == "on" } { 
         PPP_CHAP_start s $s(LCP,mode)
      }

   } elseif { $lcp(code) == "03" } {

      # -- configuration nak'd
      pktsrcWarn s "LCP configuration was NAK'd"

   } elseif { $lcp(code) == "04" } {
 
      # -- configuration rejected
      pktsrcWarn s "LCP configuration was rejected"

   } elseif { $lcp(code) == "05" } {

      pktsrcWarn s "LCP peer is terminating the link"

      # -- send terminate-ack
      LCP_send_terminate_ack s lcp

      # -- tell PPPoE we are going down
      PPPoE_terminate_service s
      
   } elseif { $lcp(code) == "09" } {

      if { "$s(LCP,echoMode)" == "on" } {
          LCP_send_echo_reply s lcp
      }
   }
}

proc LCP_renegotiate { stack } {
   upvar $stack s

   # -- mark LCP as down
   set s(LCP,state) down

   # -- send a config request
   LCP_send_config_request s

   set s(LCP,renegotiate) 1

}

proc LCP_send_config_request { stack { options "" }} {
   upvar $stack s

   if { "$options" != "" } {
      error "LCP_send_config_request: options not supported yet"
   }

   PPP_LCP_next_identifier s   
   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 01 ] (id $s(LCP,identifier))"
   set id [ format "%2.2x" $s(LCP,identifier) ]

   # -- build request with mru and magic number and auth
   set len 14
   set authOp ""
   if { "$s(PAP,mode)" == "on" } {
       set authOp "${authOp}0304c023"
       incr len 4
   } elseif { "$s(CHAP,mode)" == "on" } {

       # -- check for MSCHAP
       if { "$s(CHAP,type)" == "mschap" } {
          set authOp "${authOp}0305c22380"
       } else {
          # -- regular CHAP
          set authOp "${authOp}0305c22305"
       }
       incr len 5
   }

   set lenstr [ format "%4.4x" $len ]
   set packet "01${id}${lenstr}0104${s(LCP,mru)}0506${s(LCP,magic-number)}$authOp"

   # -- send as LCP
   PPP_send s $packet c021
}

proc LCP_send_config_ack { stack lcp } {
   upvar $stack s
   upvar $lcp l

   # - now build an options list with all the options in the lcp packet
   set optList [ LCP_build_options stack l ]

   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 02 ] (id [expr 0x$l(id)])"

   # -- build config ack for this id with this set of options
   set packet "02${l(id)}${optList}"

   # -- send as a LCP packet to PPP
   PPP_send s $packet c021
}

proc LCP_send_echo_reply { stack lcp } {
   upvar $stack s
   upvar $lcp l

   # -- build an initial empty options list
   set optList "0000"

   # - now build an echo reply with my magic number
   set optList "$s(LCP,magic-number)$l(data)"

   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 0a ] (id [expr 0x$l(id)])"

   set len [ expr [ string length $optList ] /2 ]
   incr len 4
   set lenHex [ format "%4.4x" $len ]

   # -- build config ack for this id with this set of options
   set packet "0a${l(id)}${lenHex}${optList}"

   # -- send as a LCP packet to PPP
   PPP_send s $packet c021
}

proc LCP_send_echo_request { stack data } {
   upvar $stack s

   # - now build an echo reply with my magic number
   set optList "$s(LCP,magic-number)$data"

   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 09 ] with id [expr 0x$s(LCP,identifier)], data '$data'"

   set len [ expr [ string length $optList ] /2 ]
   incr len 4
   set lenHex [ format "%4.4x" $len ]

   set id [ format "%2.2x" $s(LCP,identifier) ]

   # -- build echo request for this id with this set of options
   set packet "09${id}${lenHex}${optList}"

   # -- send as a LCP packet to PPP
   PPP_send s $packet c021
}

proc LCP_send_terminate { stack } {
   upvar $stack s

   set id [ format "%2.2x" $s(LCP,identifier) ]
   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 05 ] (id $s(LCP,identifier))"
   
   # -- build config ack for this id with this set of options
   set packet "05${id}0004"

   # -- send as a LCP packet to PPP
   PPP_send s $packet c021
}


proc LCP_send_terminate_ack { stack lcp } {
   upvar $stack s
   upvar $lcp l

   pktsrcInfo s "Sending LCP [ PPP_LCP_get_code 06 ] (id [expr 0x$l(id)])"

   # -- build config ack for this id with this set of options
   set packet "06${l(id)}0004"

   # -- send as a LCP packet to PPP
   PPP_send s $packet c021
}

proc LCP_build_options { stack op } {
   upvar $stack s
   upvar $op o

   set oplist ""
   set totalLen 0

   for { set i 0 } { $i < $o(options) } { incr i } {

       set tag    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)      
       set oplist ${oplist}${tag}${length}${data}      

       incr totalLen [ expr 0x$length ]
   }     

   # -- add 4 to include LCP header
   incr totalLen 4

   set len [ format "%4.4x" $totalLen ]

   set oplist "${len}${oplist}"

   return $oplist
}



proc PPP_LCP_start { stack { mode client }} {
   upvar $stack s
   # -- build a configure request

   set s(LCP,mode) $mode

   pktsrcInfo s "Starting LCP"
   # -- Add callbacks for all LCP packets
   Stack_add_callback s lcpType 01 LCP_client_input
   Stack_add_callback s lcpType 02 LCP_client_input
   Stack_add_callback s lcpType 03 LCP_client_input
   Stack_add_callback s lcpType 04 LCP_client_input
   Stack_add_callback s lcpType 05 LCP_client_input
   Stack_add_callback s lcpType 06 LCP_client_input
   Stack_add_callback s lcpType 07 LCP_client_input
   Stack_add_callback s lcpType 08 LCP_client_input
   Stack_add_callback s lcpType 09 LCP_client_input
   Stack_add_callback s lcpType 10 LCP_client_input
   Stack_add_callback s lcpType 11 LCP_client_input

   # -- send config request if we are acting like a client
   if { "$mode" == "client" } {
      LCP_send_config_request s
   }
}

proc PPP_LCP_terminate { stack } {
   upvar $stack s

   pktsrcInfo s "Terminating PPP link"

   LCP_send_terminate s

   set result [ PPP_LCP_get_packet s 06 packet 5000 ]

   if { "$result" != "-1" } {

      pktsrcInfo s "PPP link has been terminated"
      PPPoE_terminate_service s

   } else {

      pktsrcInfo s "no response from PPP peer for terminate request"
      PPPoE_terminate_service s
   }

   # -- reset LCP
   PPP_LCP_reset s

}

#
# -- perform a LCP ping

proc PPP_LCP_ping { stack { count 5 } } {
   upvar $stack s

   set to 0
   set response 0

   for { set i 1 } { $i <= $count } { incr i } {

      # -- send the packet
      set data [ format "%8.8x" $i]

      PPP_LCP_next_identifier s

      LCP_send_echo_request s $data

      # -- wait 5 seconds
      set result [ PPP_LCP_get_packet s 0a lcppacket 5000 ]

      if { "$result" == "-1" } {
         pktsrcWarn s "Time out waiting for LCP response"
         incr to
      } else {
         pktsrcInfo s "Received LCP echo reply with id [expr 0x$lcppacket(id)]"
         set bump 1

	 if { "$lcppacket(data)" == "$data" } {
            pktsrcInfo s "LCP data is correct"
         } else {
            pktsrcWarn s "LCP data is not correct"
            set bump 0
         }

         set myid [ format "%2.2x" $s(LCP,identifier) ]

	 if { "$lcppacket(id)" == "$myid" } {
            pktsrcInfo s "LCP identifier is correct"
         } else {
            pktsrcWarn s \
	    "LCP identifier is not correct, expecting id [ expr $s(LCP,identifier)]"
            set bump 0
         }

         # -- bump the response count if we passed all the tests
         incr response $bump
      }
   }

   return $response
}

   
proc PPP_LCP_echo_disable { stack } {
   upvar $stack s
   set s(LCP,echoMode) off
}

proc PPP_LCP_echo_enable { stack } {
   upvar $stack s
   set s(LCP,echoMode) on
}

proc PPP_LCP_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
      Event_got_packet
   }
}

#
# PPP_LCP_get_packet
#
# blocking call for receiving LCP packets
# this call blocks until the packet is received or a timeout occurs
#

proc PPP_LCP_get_packet { stack type packet timeOut } {
   upvar $stack s

   set ret -1

   # add a callback for LCP '$type' packets
   Stack_add_callback s lcpType $type PPP_LCP_generic_input

   pktsrcInfo s \
     "Waiting up to $timeOut seconds for LCP [PPP_LCP_get_code $type ] packet"

   # -- compute timeout in milli seconds
   set msec $timeOut

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the DHCP packet from the stack
      set lcpState $s(current,LCP)
      upvar #$s(current,LCP,level) $lcpState lcp
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet lcp p

      set ret 0

   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "$timeOut msec timeout waiting for [ PPP_LCP_get_code $type ] packet"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s lcpType $type PPP_LCP_generic_input

   return $ret
}






