#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ppp-ncp.tcl,v 1.5 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc PPP/NCP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc PPP_NCP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      if [ catch { PPP_NCP_decode $packet ncp } reason ] {
         pktsrcWarn s "Unable to fully parse PPP NCP packet"
      } else {
         PPP_NCP_print ncp
      }

      # -- nothing above ncp, return
      return
   }

   # -- remember this packet in the 'ncp' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s NCP ncp ]

   # -- now decode the packet into the 'ncp' variable
   if [ catch { PPP_NCP_decode $packet ncp } reason ] {
      pktsrcWarn s "Unable to fully parse PPP NCP packet"
      return
   }

   # -- check for callbacks 

   if [ info exists s(ncpType,$ncp(code)) ] {

      # -- drive any input callbacks
      foreach c $s(ncpType,$ncp(code),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   } else {
      # no callback
   }
}

#
#
# PPP_NCP_init - set up any defaults for NCP
#

proc PPP_NCP_init { stack } {
   upvar $stack s

   # -- initial NCP options
   set s(NCP,dns1)     0.0.0.0
   set s(NCP,dns2)     0.0.0.0
   set s(NCP,clientIP) 0.0.0.0

   # -- if the identifier exists already, just bump it
   if [ info exists s(NCP,identifier) ] {
      incr s(NCP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(NCP,identifier) > 255 } {
          set s(NCP,identifier) 1
       }         

   } else {

      # -- otherwise start at one
      set s(NCP,identifier) 1
   }

   set s(NCP,state) down

}

proc PPP_NCP_decode { packet NCP } {
   upvar $NCP out

   # -- parse out all the fields we find interesting ...
   set out(code)         [ string range $packet 0 1 ]
   set out(id)           [ string range $packet 2 3 ]
   set out(length)       [ string range $packet 4 7 ]

   # -- check for tags to decode   
   set len [expr [ string length $packet ] / 2 ]

   set out(options) 0

   set code [ expr 0x$out(code) ] 

   # -- decode options for those codes that support them
   if { $code <= 4 } {

      if { [expr 0x$out(length)] > $len } {
         puts "ERROR: Missing NCP data (reported [expr 0x$out(length)], found $len)"
   
      } elseif { $len > 8 } {
         PPP_NCP_decode_options $packet out
      }
   } else {

      # -- handle other codes here
      # +++joem, print proc needs to handle these separately as well

   }
}

proc PPP_NCP_decode_options { packet NCP } {
   upvar $NCP out

   set p 8

   set out(options) 0

   set count 0
   set totalLen 4

   while { 1 } {
     
      if { $totalLen >= [ expr 0x$out(length) ] } {
         break
      } 
      set tag [ string range $packet $p [expr $p + 1]]
      #puts "tag is $tag"

      incr p 2
      set length [ string range $packet $p [expr $p + 1]]
      #puts "length is $length"

      incr p 2

      set out(option,$count,tag) $tag
      set out(option,$count,length) $length

      if { "$length" == "00" } {
         puts "NCP: bad option length"
         break
      }

      set out(option,$count,data) \
          [ string range $packet $p [expr $p + 2 * 0x$length - 5 ]]

      incr p [ expr 0x$length * 2 - 4 ]

      #puts "data is $out(option,$count,data)"
      incr count

      incr totalLen [ expr 0x$length ]
   } 

   # -- set the total number of tags found
   set out(options) $count  
}

proc PPP_NCP_send { stack packet protocol } {

   upvar $stack s

   #set ppp ${protocol}${packet}

   #PPPoE_session_send s $ppp

}



proc PPP_NCP_print { NCP } {
   upvar $NCP out

   puts "NCP: ---- NCP Packet ----"
   puts "NCP: "
   puts "NCP: Code                       = [PPP_NCP_get_code $out(code)] (0x$out(code))"
   puts "NCP: Identifier                 = [ expr 0x$out(id)]"
   puts "NCP: Length                     = [ expr 0x$out(length)]"
   puts "NCP: "

   puts "NCP: ---- Options ----"
   puts "NCP:"

   if { $out(options) > 0 } {
      PPP_NCP_print_options out
   } else {
      puts "NCP: No options found"
   }

   puts "NCP:"
   puts "NCP: ---- End NCP Packet ----"

}

proc PPP_NCP_print_options { NCP } {
   upvar $NCP out

#
#         0       RESERVED
#         1       Maximum-Receive-Unit
#         3       Authentication-Protocol
#         4       Quality-Protocol
#         5       Magic-Number
#         7       Protocol-Field-Compression
#         8       Address-and-Control-Field-Compression
#

   for { set i 0 } { $i < $out(options) } { incr i } {

      # -- get the option tag in decimal

      set tag $out(option,$i,tag)
      set length $out(option,$i,length)
      set data $out(option,$i,data)

      if { "$tag" == "01" } {

         puts "NCP: IP-Addresses                = $data (deprciated?)"

      } elseif { "$tag" == "02" } {

         puts "NCP: IP-Compression Protocol     = $data"

      } elseif { "$tag" == "03" } {

         puts "NCP: IP-Address                 = [ hextoip $data ]"

#
# from rfc 1877
#
#
      } elseif { "$tag" == "81" } {

         puts "NCP: Primary-DNS-Address        = [ hextoip $data ]"

      } elseif { "$tag" == "82" } {

         puts "NCP: Primary-NBNS-Address       = [ hextoip $data ]"

      } elseif { "$tag" == "83" } {

         puts "NCP: Secondary-DNS-Address      = [ hextoip $data ]"

      } elseif { "$tag" == "84" } {

         puts "NCP: Secondary-NBNS-Address     = [ hextoip $data ]"

 
      } else {

         # -- default option display

         puts "NCP: Option Type                = (0x$tag)"
         puts "NCP: Option Length              = $length"
      }
   }
}

proc PPP_NCP_get_code { code } {
  set str "Unknown"

  set code [ expr 0x$code ]

#
# from rfc 1332
#
# only LCP codes 1 - 7 are used for NCP
#
#
#                1       Configure-Request
#                2       Configure-Ack
#                3       Configure-Nak
#                4       Configure-Reject
#                5       Terminate-Request
#                6       Terminate-Ack
#                7       Code-Reject

  if { "$code" == "1" } {
     set str  "Configure-Request"
  } elseif { "$code" == "2" } {
      set str "Configure-Ack"
  } elseif { "$code" == "3" } {
      set str "Configure-Nak"
  } elseif { "$code" == "4" } {
      set str "Configure-Reject"
  } elseif { "$code" == "5" } {
      set str "Terminate-Request"
  } elseif { "$code" == "6" } {
      set str "Terminate-Ack"
  } elseif { "$code" == "7" } {
      set str "Code-Reject"
  }

  return $str

}


proc PPP_NCP_get_option { op tag } {
   upvar $op o

   set index [ PPP_NCP_find_option o $tag ]
   if { $index > -1 } {
       return $o(option,$index,data)
   }

   return ""
}

proc PPP_NCP_add_option { op tag length value } {
   upvar $op o

   if ![ info exists o(options) ] {
     set o(options) 0
   } 

   set index $o(options)
   set o(option,$index,length) $length
   set o(option,$index,tag) $tag
   set o(option,$index,data) $value

   # -- bump the option count
   incr o(options)
}

proc PPP_NCP_update_option { op tag value } {
   upvar $op o

   set index [ PPP_NCP_find_option o $tag ]
   if { $index > -1 } {
       # puts "updating option $tag with $value at index $index"
       set o(option,$index,data) $value

   }
}

proc PPP_NCP_find_option { op tag { from 0 }} {
   upvar $op o

   set oplist ""
   set totalLen 0

   for { set i $from } { $i < $o(options) } { incr i } {

       if { "$tag" == "$o(option,$i,tag)" } {
           return $i
       }
   }
   return -1
}

proc NCP_client_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the NCP packet from the stack
   set ncpState $s(current,NCP)
   upvar #$s(current,NCP,level) $ncpState ncp

   pktsrcInfo s "Received NCP [ PPP_NCP_get_code $ncp(code) ]"

   if { $ncp(code) == "01" } {

      # -- got configuration request, just sent ack if client
      if { "$s(LCP,mode)" == "client" } {
          NCP_send_config_ack s ncp
      } else {

          # -- see if the options match
          set addr [ PPP_NCP_get_option ncp 03 ]
          set dns1 [ PPP_NCP_get_option ncp 81 ]
          set dns2 [ PPP_NCP_get_option ncp 83 ]

          set sendNak 0

          # -- check address
          if { "$addr" != "" } {
                 pktsrcInfo s "requested IP Address [ hextoip $addr]"
	         if {   "$addr" != "[ iptohex $s(NCP,clientIP)]" }  {
                    pktsrcInfo s "rejecting IP Address"
                    set sendNak 1
                 }
	  }

          if { "$dns1" != "" } {
                 pktsrcInfo s "requested Primary DNS Server [ hextoip $dns1]"
	         if {   "$dns1" != "[ iptohex $s(NCP,dns1)]" }  {
                    pktsrcInfo s "rejecting Primary DNS Server"
                    set sendNak 1
                 }
	  }

          if { "$dns2" != "" } {
                 pktsrcInfo s "requested Secondary DNS Server [ hextoip $dns2]"
	         if {   "$dns2" != "[ iptohex $s(NCP,dns2)]" }  {
                    pktsrcInfo s "rejecting Secondary DNS Server"
                    set sendNak 1
                 }
	  }

          if { "$sendNak" == "1" } {
               # nak this configuration
               pktsrcInfo s "not accepting NCP peer configuration"
               NCP_send_config_nack s ncp

          } else {
               pktsrcInfo s "accepting NCP peer configuration"
               NCP_send_config_ack s ncp

               # -- send my configuration request
               NCP_send_config_request s
          }

      }



   } elseif { $ncp(code) == "02" } {

      pktsrcInfo s "NCP configuration was accepted"
      set s(NCP,state) up

   } elseif { $ncp(code) == "03" } {

      # -- configuration nak'd
      pktsrcWarn s "NCP configuration was NAK'd"

   } elseif { $ncp(code) == "04" } {
 
      # -- configuration rejected
      pktsrcWarn s "NCP configuration was rejected"
   }
}

proc NCP_send_config_request { stack { options "" }} {
   upvar $stack s

   if { "$options" != "" } {
      error "NCP_send_config_request: options not supported yet"
   }

   # -- build request with address
   pktsrcInfo s "Sending NCP [ PPP_NCP_get_code 01 ] with IP Address [hextoip $s(ip)]"

   # -- build ncp config request
   set packet "0101000a0306$s(ip)"

   # -- send as NCP
   PPP_send s $packet 8021
}

proc NCP_send_config_ack { stack ncp } {
   upvar $stack s
   upvar $ncp l

   # - now build an options list with all the options in the NCP packet
   set optList [ NCP_build_options l ]

   pktsrcInfo s "Sending NCP [ PPP_NCP_get_code 02 ]"

   # -- build config ack for this id with this set of options
   set packet "02${l(id)}${optList}"

   # -- send as a NCP packet to PPP
   PPP_send s $packet 8021
}

proc NCP_send_config_nack { stack ncp } {
   upvar $stack s
   upvar $ncp l

   # -- for config nack, use a separate option list
   PPP_NCP_add_option newOpt 03 06 [ iptohex $s(NCP,clientIP) ]
   PPP_NCP_add_option newOpt 81 06 [ iptohex $s(NCP,dns1) ]
   PPP_NCP_add_option newOpt 83 06 [ iptohex $s(NCP,dns2) ]

   #if { "$s(NCP,dns1)" != "0.0.0.0" } {
   #   puts ">>> Adding dns1 option $s(NCP,dns1)"
   #
   #}

   #if { "$s(NCP,dns2)" != "0.0.0.0" } {
   #   puts ">>> Adding dns2 option $s(NCP,dns2)"
   #
   #  }

   # - now build an options list with all the options in the NCP packet
   set optList [ NCP_build_options newOpt ]

   if { "$s(NCP,dns2)" != "0.0.0.0" } {
      pktsrcInfo s "Sending NCP [ PPP_NCP_get_code 03 ], use IP Address $s(NCP,clientIP), primary DNS server $s(NCP,dns1), secondary DNS server $s(NCP,dns2)"
   } else {
      pktsrcInfo s "Sending NCP [ PPP_NCP_get_code 03 ], use IP Address $s(NCP,clientIP), primary DNS server $s(NCP,dns1)"
   }

   # -- build config ack for this id with this set of options
   set packet "03${l(id)}${optList}"

   # -- send as a NCP packet to PPP
   PPP_send s $packet 8021
}

proc NCP_build_options { op } {
   upvar $op o


   set oplist ""
   set totalLen 0

   for { set i 0 } { $i < $o(options) } { incr i } {

       # puts "build option $i"
       set tag    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)      
       set oplist ${oplist}${tag}${length}${data}      
       # puts "build option $i, $tag, $length, $data"
       incr totalLen [ expr 0x$length ]
   }     

   # -- add 4 to include NCP header
   incr totalLen 4

   set len [ format "%4.4x" $totalLen ]

   # puts "option length is $totalLen"

   set oplist "${len}${oplist}"

   return $oplist
}



proc PPP_NCP_start { stack { mode client }} {
   upvar $stack s
   # -- build a configure request

   pktsrcInfo s "Starting NCP/IPCP"

   # -- Add callbacks for all NCP packets
   Stack_add_callback s ncpType 01 NCP_client_input
   Stack_add_callback s ncpType 02 NCP_client_input
   Stack_add_callback s ncpType 03 NCP_client_input
   Stack_add_callback s ncpType 04 NCP_client_input
   Stack_add_callback s ncpType 05 NCP_client_input
   Stack_add_callback s ncpType 06 NCP_client_input
   Stack_add_callback s ncpType 07 NCP_client_input

   # -- send config request
   if { $mode == "client" } {
      NCP_send_config_request s
   }
}












