#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ppp-pap.tcl,v 1.6 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc PPP/PAP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc PPP_PAP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {
      PPP_PAP_decode $packet pap
      PPP_PAP_print pap

      # -- nothing above pap, return
      return
   }

   # -- remember this packet in the 'pap' variable. This allows other
   # callbacks to get access to the decodes at any Tcl stack level

   eval [ Stack_save_packet s PAP pap ]
   
   # -- now decode the packet into the 'pap' variable
   PPP_PAP_decode $packet pap

   # -- check for callbacks 

   if [ info exists s(papType,$pap(code)) ] {

      # -- drive any input callbacks
      foreach c $s(papType,$pap(code),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }
   } else {
      # no callback
   }
}

#
#
# PPP_PAP_init - set up any deaults for PAP
#

proc PPP_PAP_init { stack } {
   upvar $stack s

   # -- initial PAP options

   set s(PAP,user) joe
   set s(PAP,password) hello
   set s(PAP,mode) off 


   # -- if the identifier exists already, just bump it
   if [ info exists s(PAP,identifier) ] {
      incr s(PAP,identifier)

       # -- wrap after 255 since identifier is 1 byte field
       if { $s(PAP,identifier) > 255 } {
          set s(PAP,identifier) 1
       }         

   } else {

      # -- otherwise start at one
      set s(PAP,identifier) 1
   }

   set s(PAP,state) down
}

proc PPP_PAP_decode { packet PAP } {
   upvar $PAP out

   # -- parse out all the fields we find interesting ...
   set out(code)         [ string range $packet 0 1 ]
   set out(id)           [ string range $packet 2 3 ]
   set out(length)       [ string range $packet 4 7 ]

   # -- check for tags to decode   
   set len [expr [ string length $packet ] / 2 ]

   set out(options) 0
   set code [ expr 0x$out(code) ] 

   if { "$code" == "1" } {
      set out(peer-id-len)       [ string range $packet 8 9 ]
      set len [ expr 0x$out(peer-id-len) ]
      set myend [ expr 10 + ($len * 2 ) - 1]
      set out(peer-id) [ string range $packet 10 $myend ]

      incr myend
      set end2 [ expr $myend + 1]
      set out(password-len)       [ string range $packet $myend $end2 ]
      set len [ expr 0x$out(password-len) ]
      incr end2
      set myend [ expr $end2 + ($len * 2 ) - 1]
      set out(password) [ string range $packet $end2 $myend ]
   }
}

proc PPP_PAP_send { stack packet protocol } {

   upvar $stack s

   #set ppp ${protocol}${packet}

   #PPPoE_session_send s $ppp

}



proc PPP_PAP_print { PAP } {
   upvar $PAP out

   set code [ expr 0x$out(code) ]

   puts "PAP: ---- PAP Packet ----"
   puts "PAP: "
   puts "PAP: Code                       = [PPP_PAP_get_code $out(code)] (0x$out(code))"
   puts "PAP: Identifier                 = [ expr 0x$out(id)]"
   puts "PAP: Length                     = [ expr 0x$out(length)]"

   set code [ expr 0x$out(code) ] 

   if { "$code" == "1" } {
      set user [ pktsrc ascii2text $out(peer-id) ]
      set pass [ pktsrc ascii2text $out(password) ]
      puts "PAP: Peer ID                    = $user"
      puts "PAP: Password                   = $pass"
   }

   puts "PAP:"
   puts "PAP: ---- End PAP Packet ----"

}

proc PPP_PAP_get_code { code } {
  set str "Unknown"

  set code [ expr 0x$code ]

#
#         1       Authenticate-Request
#         2       Authenticate-Ack
#         3       Authenticate-Nak
#

  if { "$code" == "1" } {
     set str  "Authenticate-Request"
  } elseif { "$code" == "2" } {
      set str "Authenticate-Ack"
  } elseif { "$code" == "3" } {
      set str "Authenticate-Nak"
  }

  return $str

}

proc PPP_PAP_get_auth { type } {
  set str "Unknown"

  if { "$type" == "c22305" } {
     set str  "CHAP"
  } elseif { "$type" == "c22380" } {
     set str  "MSCHAP"
  } elseif { "$type" == "c023" } {
      set str "PAP"
  }

  return $str

}

proc PAP_set_user { stack user password } {
   upvar $stack s

   set s(PAP,user) $user
   set s(PAP,password) $password

   pktsrcInfo s "Setting PPP PAP user to $user with password $password"

}

proc PAP_get_user { stack } {
   upvar $stack s
   return $s(PAP,user)
}

proc PAP_get_password { stack } {
   upvar $stack s
   return $s(PAP,password)
}


proc PAP_client_input { packet stack { options none }} {
   upvar $stack s

   # -- get the contents of the PAP packet from the stack

   set papState $s(current,PAP)
   upvar #$s(current,PAP,level) $papState pap

   pktsrcInfo s "Received PAP [ PPP_PAP_get_code $pap(code) ]"

   if { $pap(code) == "01" } {

      set user [ pktsrc ascii2text $pap(peer-id) ]
      set pass [ pktsrc ascii2text $pap(password) ]

      if { "$user" == "$s(PAP,user)" &&  "$pass" == "$s(PAP,password)" } {
          pktsrcInfo s "PAP authentication OK, user '$user', password '$pass'"
          PAP_send_config_ack s pap
          set s(PAP,state) up
      } else {
          pktsrcInfo s "PAP authentication failed, user '$user', password '$pass'"
          PAP_send_config_nack s pap
          set s(PAP,state) down
      }



   } elseif { $pap(code) == "02" } {
      pktsrcInfo s "PAP authentication was accepted"

   } elseif { $pap(code) == "03" } {

      # -- configuration nak'd
      pktsrcWarn s "PAP authentication was NAK'd"

   }
}

proc PAP_send_auth_request { stack { options "" }} {
   upvar $stack s

   if { "$options" != "" } {
      error "PAP_send_config_request: options not supported yet"
   }

   pktsrcInfo s "Sending PAP [ PPP_PAP_get_code 01 ]"

   # -- build request with mru and magic number
   set packet "0101000d0104${s(LCP,mru)}0506$s(LCP,magic-number)"

   # -- send as PAP
   PPP_send s $packet c023
}

proc PAP_send_config_ack { stack pap } {
   upvar $stack s
   upvar $pap l

   # - now build an options list with all the options in the lcp packet
   # set optList [ PAP_build_options stack l ]

   pktsrcInfo s "Sending PAP [ PPP_PAP_get_code 02 ]"

   # -- build config ack for this id
   set packet "02${l(id)}000500"

   # -- send as a PAP packet to PPP
   PPP_send s $packet c023
}

proc PAP_send_config_nack { stack pap } {
   upvar $stack s
   upvar $pap l

   pktsrcInfo s "Sending PAP [ PPP_PAP_get_code 03 ]"

   # -- build config nack for this id with this set of options
   set packet "03${l(id)}0004"

   # -- send as a PAP packet to PPP
   PPP_send s $packet c023
}

proc PAP_build_options { stack op } {
   upvar $stack s
   upvar $op o

   set oplist ""
   set totalLen 0

   for { set i 0 } { $i < $o(options) } { incr i } {

       set tag    $o(option,$i,tag)
       set length $o(option,$i,length)
       set data   $o(option,$i,data)      
       set oplist ${oplist}${tag}${length}${data}      

       incr totalLen [ expr 0x$length ]
   }     

   # -- add 4 to include LCP header
   incr totalLen 4

   set len [ format "%4.4x" $totalLen ]
   set oplist "${len}${oplist}"

   return $oplist
}



proc PPP_PAP_start { stack { mode client } } {
   upvar $stack s
   # -- build a configure request

   pktsrcInfo s "Starting PAP"
   # -- Add callbacks for all LCP packets
   Stack_add_callback s papType 01 PAP_client_input
   Stack_add_callback s papType 02 PAP_client_input
   Stack_add_callback s papType 03 PAP_client_input

   # -- send config request
   if { $mode == "client" } {
      PAP_send_auth_request s
   }
}







