#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: ppp.tcl,v 1.6 2002/11/26 19:20:20 joe Exp $
#
#
# pktsrc PPP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc PPP_input { packet stack { options none }} {
   upvar $stack s

   set port $s(port)
   PPP_decode $packet ppp


   if { $options == "print" } {
      PPP_print ppp
   } else {

     # -- I should hide this in a save proto
     set s(current,PPP) ppp
     set s(current,PPP,level) [info level]
   }

   # -- check for callbacks 

   if [ info exists s(pppType,$ppp(protocol)) ] {

      # -- drive any input callbacks
      foreach c $s(pppType,$ppp(protocol),callbacks) {
          # -- send the whole packet
          #$c $ppp(data) s $options
         if [ catch { $c $ppp(data) s $options } reason ] {
            puts "PKTSRC-ERROR: input callback '$c' failed, $options"
            puts "PKTSRC-ERROR: $reason"
         }
          # $s(pppType,$ppp(protocol),input) $ppp(data) s $options
      }
   } else {
      # no callback
   }
}

#
#
# PPP_init - set up any deaults for PPP
#

proc PPP_init { stack } {
   upvar $stack s


}

proc PPP_decode { packet PPP } {
   upvar $PPP out

   # -- parse out all the fields we find interesting ...
   set out(protocol)       [ string range $packet 0 3 ]

   set out(data) [ string range $packet 4 end ]
}

proc PPP_send { stack packet protocol } {

   upvar $stack s

   set ppp ${protocol}${packet}

   PPPoE_session_send s $ppp

}



proc PPP_print { PPP } {
   upvar $PPP out

   puts "PPP: ---- PPP Packet ----"
   puts "PPP: "
   puts "PPP: Protocol                   = [PPP_get_protocol $out(protocol)] (0x$out(protocol))"
   puts "PPP: "
}


proc PPP_get_protocol { code } {
  set str "Unknown "

  if { "$code" == "c021" } {
     set str  "LCP"
  } elseif { "$code" == "0021" } {
     set str  "IP"
  } elseif { "$code" == "8021" } {
      set str "NCP/IPCP"
  } elseif { "$code" == "c023" } {
      set str "PAP"
  } elseif { "$code" == "c223" } {
      set str "CHAP"
  }
  return $str

}

proc PPP_client { stack } {
  upvar $stack s

  PPP_init s
  pktsrcInfo s "Starting PPP client"
  PPP_LCP_start s 

}

proc PPP_server { stack } {
  upvar $stack s

  if [ info exists s(PPP,server,running) ] {

     # -- already running
     return

  } else {

     set s(PPP,server,running) 1

     PPP_init s
     pktsrcInfo s "Starting PPP server mode"
     PPP_LCP_start s server
  }
}

proc PPP_set_user { stack user password } {
   upvar $stack s
 
   # -- set the users for both PAP and CHAP   
   PAP_set_user s $user $password
   CHAP_set_user s $user $password

}





