#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2001 QAcafe http://qacafe.com
#
# $Id: rip.tcl,v 1.5 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc RIP
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc RIP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      RIP_decode $packet rip
      RIP_print rip
      return

   } else {

      # -- remember this packet in the 'rip' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s RIP rip ]
   
      # -- now decode the packet into the 'rip' variable
      RIP_decode $packet rip

   }

   # -- check for callbacks at the RIP level
   if [ info exists s(ripType,$rip(command)) ] {

      # -- drive any input callbacks
      foreach c $s(ripType,$rip(command),callbacks) {
          # -- send the whole packet
          $c $packet s $options
      }

   } else {
      # no callback
   }
}

#
#
# RIP_init - set up any deaults for RIP
#

proc RIP_init { stack } {
   upvar $stack s

   # -- set the default version
   set s(RIP,version) 1
}


proc RIP_set_version { stack version } {
   upvar $stack s

   if { $version == 1 } {
 
      # -- set the version
      set s(RIP,version) $version

   } elseif { $version == 2 } {

      # -- set the version and add a mac for RIP multicast
      set s(RIP,version) $version
      ARP_add_entry s [iptohex 224.0.0.9] 01005e000009

   } else {
     error "pktsrc: unknown RIP version '$version'"
   }
}


proc RIP_decode { packet RIP } {
   upvar $RIP out

   # -- parse out all the fields we find interesting ...

   set out(command)        [ string range $packet 0 1 ]
   set out(version)        [ string range $packet 2 3 ]
   set out(mbz1)           [ string range $packet 4 7 ]
   set out(family)         [ string range $packet 8 11 ]
   set out(mbz2)           [ string range $packet 12 15 ]

   if { $out(command) == "02" } {
      RIP_decode_response $packet out
   }

   if { $out(command) == "01" } {
      RIP_decode_response $packet out
   }
}

proc RIP_decode_response { packet rip } {
   upvar $rip out

   set p 8

   set out(total) 0

   while { 1 } {
      set tuple [ string range $packet $p end ]
      set len [ string length $tuple ]
      if { $len >= 40 } {
         incr p 40
         incr out(total)
         set out(route,$out(total)) $tuple
      } else {
         # -- not enough data for full route
         break
      }
   }
}
       

proc RIP_print { RIP } {
   upvar $RIP out

   puts "RIP: ---- RIP Packet ----"
   puts "RIP: "
   puts "RIP: Command                    = [RIP_command $out(command)] (0x$out(command))"
   puts "RIP: Version                    = [expr 0x$out(version)]"
   puts "RIP: Family                     = 0x$out(family)"

   if { $out(command) == "02" } {
       RIP_print_response out
   }

   if { $out(command) == "01" } {
       RIP_print_response out
   }

   puts "RIP:"
   puts "RIP: ---- End RIP Packet ----"
}

proc RIP_print_response { rip } {
   upvar $rip out

   puts "RIP: ----  Routes  ----"
   puts "RIP: Number of Entries          = $out(total)" 
 
   for { set i 1 } { $i <= $out(total) } { incr i } { 
       if [ info exists out(route,$i) ] {
           RIP_print_route $out(route,$i) $out(version)
       } else {
           break
       }
   }
}

proc RIP_print_route { route { version 1 }} {
   RIP_decode_route $route out

   puts "RIP:"   
   puts "RIP: Address                    = [hextoip $out(address)]"
   if { $version == 2 } {
     puts "RIP: Tag                        = 0x$out(tag)"
     puts "RIP: Mask                       = [hextoip $out(mask)]"
     puts "RIP: Next-hop                   = [hextoip $out(nexthop)]"
   }
   puts "RIP: Metric                     = [expr 0x$out(metric)]"

}

proc RIP_decode_route { route output } {
   upvar $output out

   set out(family)  [ string range $route 0 3 ]
   set out(tag)     [ string range $route 4 7 ]
   set out(address) [ string range $route 8 15 ]
   set out(mask)    [ string range $route 16 23 ]
   set out(nexthop) [ string range $route 24 31 ]
   set out(metric)  [ string range $route 32 39 ]
}


proc RIP_command { type } {

   if { $type == "01" } {
      set str "Request"
   } elseif { $type == "02" } {
      set str "Reply"
   } elseif { $type == "03" } {
      set str "Trace On"
   } elseif { $type == "04" } {
      set str "Trace Off"
   } elseif { $type == "05" } {
      set str "Reserved (SUN)"
   } else {
      set str "Unknown RIP Command ($type)"
   }

   return $str
}


#
#

proc RIP_send_request { stack destIP srcPort } {
   upvar $stack s

   set packet [ RIP_build_request $s(RIP,version) ]
   pktsrcInfo s "Sending RIPv$s(RIP,version) request to ${destIP}:520"
   RIP_send s $packet $destIP $srcPort
}

proc RIP_send_route { stack dest mask gateway metric } {
   upvar $stack s

   set packet [ RIP_build_annoucement $s(RIP,version) $dest $mask $gateway $metric ]

   if { "$s(RIP,version)" == "1" } {
       set updatedest 255.255.255.255
   } else {
       set updatedest 224.0.0.9
   }

   pktsrcInfo s "Sending RIPv$s(RIP,version) reply for $dest to $updatedest:520"
   RIP_send s $packet $updatedest 520

}
proc RIP_send_route_list { stack list } {
   upvar $stack s
   upvar $list r


   set dest $r(1,route)
   set mask $r(1,mask)
   set gateway $r(1,nexthop)
   set metric $r(1,metric)

   # -- build the header and first route
   set packet [ RIP_build_annoucement $s(RIP,version) $dest $mask $gateway $metric ]

   # -- add additional routes
   for { set i 2 } { $i <= $r(count) } { incr i } {

     set dest $r($i,route)
     set mask $r($i,mask)
     set gateway $r($i,nexthop)
     set metric $r($i,metric)

     set packet [ format "%s00020000%s%s%s%s" $packet  \
       [iptohex $dest ] [ iptohex $mask ] [ iptohex $gateway ] \
       [ format "%8.8x" $metric ] ]

   }

   if { "$s(RIP,version)" == "1" } {
       set updatedest 255.255.255.255
   } else {
       set updatedest 224.0.0.9
   }

   pktsrcInfo s "Sending RIPv$s(RIP,version) reply with $r(count) routes to $updatedest:520"

   RIP_send s $packet $updatedest 520

}

proc RIP_build_request { version } {
   set packet [ format "01%2.2d00000000000000000000000000000000000000000010" $version ]
 
   # -- return new rip request
   return $packet
}

proc RIP_build_annoucement { version dest mask nexthop metric } {

   set packet [ format "02%2.2d000000020000%s%s%s%s" $version  \
     [iptohex $dest ] [ iptohex $mask ] [ iptohex $nexthop ] \
     [ format "%8.8x" $metric ] ]
 
   # -- return new rip request
   return $packet
}
 
#
# -- send an RIP packet

proc RIP_send { stack packet dest srcPort } {
   upvar $stack s

   UDP_send s $packet [iptohex $dest ] 520 $srcPort
}

proc RIP_client { stack } {
   upvar $stack s

   # -- get a free UDP port
   set port [ UDP_get_client_port s ]

   # -- get the port number in hex, since callbacks are hex based
   set portHex [ format "%4.4x" $port ]

   # -- add a new callback for this UDP port number
   Stack_add_callback s udpPort $portHex RIP_input

   # -- retunr the decimal port number
   return $port
}

#
# RIP helper procs
#

#
# RIP_generic_input
#
# generic callback for catching RIP packets
#

proc RIP_generic_input { packet stack { options none }} {
   upvar $stack s

   # -- trigger the packet event
   if { "$options" != "print" } {
      Event_got_packet
   }
}

#
# RIP_get_packet
#
# blocking call for receiving RIP packets
# this call blocks until the packet is received or a timeout occurs
#

proc RIP_get_packet { stack type packet timeout } {
   upvar $stack s

   set ret -1

   # add a callback for RIP '$type' packets
   Stack_add_callback s ripType $type RIP_generic_input

   pktsrcInfo s \
     "Waiting up to $timeout milli seconds for [RIP_command $type ] packet"

   # -- compute timeout in milli seconds
   set msec $timeout

   # -- now wait for the packet or timeout
   set result [ Event_wait_packet $msec  ]

   #
   # result = 2 -> got packet
   #

   if { "$result" == "2" } {

      # -- get the contents of the RIP packet from the stack
      set ripState $s(current,RIP)
      upvar #$s(current,RIP,level) $ripState rip
      upvar $packet p

      # -- copy the packet to pass back to the user
      pktsrc_copy_packet rip p

      set ret 0


   #
   # result = 1 -> timeout
   #

   } elseif { "$result" == "1" } {

      
      pktsrcInfo s \
         "$timeout second timeout waiting for RIP [ RIP_command $type ] packet"

      set ret -1

   } else {

      set ret -1
   }

   # -- removing the callback
   Stack_remove_callback s ripType $type RIP_generic_input

   return $ret
}


#
# RIP_route_table
#
#

proc RIP_rtable_clear { stack } {
   upvar $stack s

   foreach r [ array names s RIP,rtable,* ] {
      set parts [ split $r , ]

      set route [ lindex $parts 2 ]
      set mask [ lindex $parts 3 ]
      set nexthop [ lindex $parts 4 ]

      RIP_rtable_delete s $route $mask $nexthop
   }
}

proc RIP_rtable_add { stack route mask nexthop metric } {
   upvar $stack s
   set s(RIP,rtable,$route,$mask,$nexthop) $metric

}

proc RIP_rtable_delete { stack route mask nexthop } {
   upvar $stack s
   unset s(RIP,rtable,$route,$mask,$nexthop)

}

#
# Send RIP replys for all routes in the RIP routing table
#

proc RIP_rtable_update { stack } {
   upvar $stack s

   pktsrcInfo s "Sending RIP update for all routes"

   set route(count) 0

   foreach r [ lsort [ array names s RIP,rtable,* ]] {

      incr route(count)

      set parts [ split $r , ]
      set dest [ lindex $parts 2 ]
      set mask [ lindex $parts 3 ]
      set nexthop [ lindex $parts 4 ]

      set metric $s(RIP,rtable,$dest,$mask,$nexthop)

      set route($route(count),route)   $dest
      set route($route(count),mask)    $mask
      set route($route(count),nexthop) $nexthop
      set route($route(count),metric)  $metric


      if { $route(count) == "24" } {
         RIP_send_route_list s route
         set route(count) 0
      } 
        
   }

   if { "$route(count)" > 0 } {
      RIP_send_route_list s route
   }
}

proc RIP_router { stack mode } {
   upvar $stack s

   if { "$mode" == "on" } {

       if { [ info exists s(RIP,running) ] } {

        # -- RIP is already running
        pktsrcWarn s "RIP is already running"

     } else {

        pktsrcInfo s "Enabling RIPv$s(RIP,version) router"

        # -- send updates now
        RIP_rtable_update s


        # -- start timer
        set s(RIP,running) [ after 30000 "
          $s(getStack)
          RIP_rtable_update $s(getStackName)
        "]

     }


   } else {

       if { [ info exists s(RIP,running) ] } {

        # -- turn off updates     
        after cancel $s(RIP,running)
        unset s(RIP,running)

        pktsrcInfo s "Disabling RIPv$s(RIP,version) router"

     }
   }
}






