#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# 2002 QAcafe http://qacafe.com
#
# $Id: upnp.tcl,v 1.2 2002/11/15 13:23:54 joe Exp $
#
#
# pktsrc UPnP (SSDP)
#
# ----------------------------------------------------------------------------

# -- this is part of the pktsrc package
package provide pktsrc 1.3

proc SSDP_input { packet stack { options none }} {
   upvar $stack s

   if { $options == "print" } {

      SSDP_decode $packet ssdp
      SSDP_print ssdp
      return

   } else {

      # -- remember this packet in the 'ssdp' variable. This allows other
      # callbacks to get access to the decodes at any Tcl stack level

      eval [ Stack_save_packet s SSDP ssdp ]
   
      # -- now decode the packet into the 'ssdp' variable
      SSDP_decode $packet ssdp

   }

   # -- check for callbacks at the RIP level
   #if [ info exists s(ssdpType,$rip(command)) ] {
   #
   #   # -- drive any input callbacks
   #   foreach c $s(ripType,$rip(command),callbacks) {
   #       # -- send the whole packet
   #       $c $packet s $options
   #   }
   #
   #} else {
   #   # no callback
   #}
}

#
#
# SSDP_init - set up any deaults for SSDP
#

proc SSDP_init { stack } {
   upvar $stack s

   ARP_add_entry s [iptohex 239.255.255.250 ] 01005e7ffffa

}


proc SSDP_decode { packet SSDP } {
   upvar $SSDP out

   # -- just treat everything as one text string for now
   set out(data)        [ string range $packet 0 end ]

}

proc SSDP_print { SSDP } {
   upvar $SSDP out

   puts "SSDP: ---- SSDP Packet ----"
   puts "SSDP: "

   # -- display everything as one text string
   puts "[pktsrc ascii2text $out(data) ]"

   puts "SSDP:"
   puts "SSDP: ---- End SSDP Packet ----"
}

proc SSDP_send_search { stack } {
   upvar $stack s

   set command(1) "M-SEARCH * HTTP/1.1"
   set command(2) "HOST:239.255.255.250:1900"
   set command(3) "MAN:\"ssdp:discover\""
   set command(4) "MX:3"
   #set command(5) "ST:"
   # set command(5) "ST:urn:schemas-upnp-org:device:InternetGatewayDevice:1"
   set command(5) "ST:urn:schemas-upnp-org:device-1-0:"
   #set command(5) "ST:upnp:rootdevice"
   # set command(5) "ST:ssdp:all"
   # set command(5) "ST:urn:schemas-upnp-org:service:WANIPConnection:1"
   set lines 5

   set commandhex ""

   for { set i 1 } { $i <= $lines } { incr i } {
       set hex [ pktsrc text2ascii $command($i) ]
       set commandhex "${commandhex}${hex}0d0a"
   }

   # -- add trailing carriage return
   set commandhex "${commandhex}0d0a"

   # -- adjust TTL
   set ttl $s(IPv4,ttl)
   set s(IPv4,ttl) 04
   UDP_send s $commandhex [iptohex 239.255.255.250 ] 1900 1901

   # -- set the ttl back
   set s(IPv4,ttl) $ttl

}





